import n2v 
import psi4
import pytest
import numpy as np

@pytest.fixture
def ine():
    Ne = psi4.geometry("""
    0 1
    Ne
    symmetry c1
    units bohr
    """)
    # Perform Calculation
    wfn = psi4.properties("ccsd/aug-cc-pvtz", molecule=Ne, return_wfn=True, properties=['dipole'])[1]
    # Initialize inverter object. 
    ine = n2v.Inverter.from_wfn(wfn)

    return ine

def test_zmp(ine):
    ine.invert("zmp", opt_max_iter=200, opt_tol=1e-7, zmp_mixing=1, 
            lambda_list=np.linspace(10, 1000, 20), guide_components="fermi_amaldi")

    assert np.isclose( ine.eigvecs_a[:5].all(),  
                       np.array([-30.96371089,  -1.64633607,  -0.78780286,  -0.78780286, -0.78780286]).all() ) 
    assert np.isclose( np.diag(ine.Da)[:5].all(), 
                       np.array([0.9520639 , 0.33023202, 0.03466103, 0.13804668, 0.16689393]).all())

    
def test_wuyang(ine):
    
    ine.invert("WuYang", guide_components="fermi_amaldi")

    assert np.isclose( ine.eigvecs_a[:5].all(),  
                        np.array([-30.9112617 ,  -1.59602304,  -0.74201369,  -0.74201369, -0.74201369]).all() )
    assert np.isclose( np.diag(ine.Da)[:5].all(),
                       np.array([0.95298269, 0.32446633, 0.03453696, 0.14332516, 0.16706669]).all() )


def test_pedeco(ine):

    ine.invert("PDECO", opt_max_iter=200, guide_components="fermi_amaldi", gtol=1e-6)

    assert np.isclose( ine.eigvecs_a[:5].all(),  
                       np.array([-30.90838995,  -1.59752788,  -0.74329655,  -0.74328871, -0.74328082]).all() )
    assert np.isclose( np.diag(ine.Da)[:5].all(),
                       np.array([0.95291707, 0.32472977, 0.03449241, 0.1432748 , 0.16698491]).all() )

def test_oucarter(ine):

    x = np.linspace(-5, 10, 1501)
    y = [0]
    z = [0]
    grid, shape = ine.eng.grid.generate_grid(x, y, z)

    v = ine.invert("OC", vxc_grid=grid, opt_max_iter=21, frac_old=0.9, init='SCAN')

    assert np.isclose( ine.eigvecs_a[:5].all(),  
                       np.array([-31.02826629,  -1.7038403 ,  -0.85493085,  -0.85493085, -0.85493085]).all() )
    assert np.isclose( np.diag(ine.Da)[:5].all(),
                       np.array([0.94579632, 0.32844699, 0.03465013, 0.1395006 , 0.16976739]).all() )
    assert np.isclose( v[0].all(), 
                       np.array([-16.83962961,  -1.02552792,  -0.5309617 ,  -0.27295131, -0.23428725]).all() )

def test_mrks(ine):

    mol = ine.eng.mol
    psi4.set_options({'opdm' : True,
                      'tpdm' : True, 
                      'dft_spherical_points' : 50,
                      'dft_radial_points'    : 50, })

    wfn = psi4.properties('detci/cc-pcvdz', return_wfn=True, molecule=mol, properties=['dipole'])[1]
    ine = n2v.Inverter.from_wfn(wfn)
    x = np.linspace(-5, 10, 1501)
    y = [0]
    z = [0]
    grid, shape = ine.eng.grid.generate_grid(x, y, z)
    ine.invert('mRKS', vxc_grid=grid, opt_max_iter=30, frac_old=0.8, init='scan')

    assert np.isclose( ine.eigvecs_a[:5].all(),  
                       np.array([-30.0960936 ,  -1.60549235,  -0.74989818,  -0.74989818, -0.74989818]).all() )
    assert np.isclose( np.diag(ine.Da)[:5].all(),
                       np.array([0.99495901, 0.24875539, 0.34138226, 0.46664006, 0.46664006]).all() )
    assert np.isclose( ine.grid_vxc.all(), 
                       np.array([-7.23643144, -1.05559238, -0.47135156, -0.28189798, -0.20578139]).all() )





"""
mrks.py

Functions associated with mrks inversion
"""

import numpy as np
from opt_einsum import contract
import psi4
import time

class MRKS():
    """
    Wavefunction to KS_potential method based on
    [1] [PRL 115, 083001 (2015)],
    [2] [J. Chem. Phys. 146, 084103 (2017)].

    The XC potential is calculated on the grid
    instead of on the potential basis set.
    Whereas, the guide potential is still used and
    plays the role of v_hartree.
    And because of this, the grid for vxc for output
    has to be specified beforehand.
    
    For CIWavefunction as input, make sure to turn on
    option opdm and tpdm:
        psi4.set_options({
            "opdm": True,
            "tpdm": True,
            'REFERENCE': "RHF"
            })

    Attributes:
    -----------
    Vpot: psi4.core.VBase
        V_potential that contains the info of DFT spherical grid.
    npoint_DFT: int
        number of points for DFT spherical grid.
    vxc_hole_WF: np.ndarray
        vxc_hole_WF on spherical grid. This is stored
        because the calculation of this takes most time.
    """
    vxc_hole_WF = None

    def _vxc_hole_quadrature(self, grid_info=None, atol=1e-5, atol1=1e-4, Vpot=None):
        """
        Calculating v_XC^hole in [1] (15) using quadrature
        integral on the default DFT spherical grid.

        Side note: the calculation is actually quite sensitive to atol/atol1
        i.e. the way to handle singularities. Play with it if you
        have a chance.
        The current model is:
        R2 = |r1 - r2|^2
        R2[R2 <= atol] = infinity (so that 1/R = 0)
        R2[atol < R2 < atol1] = atol1

        When self.wfn.name() == CIWavefunction, opdm and tpdm are used
        for calculation;
        when self.wfn.name() == RHF, actually a simplified version (i.e.
        directly using exact HF exchange densities) is implemented. In other
        words, this part can be much quicker and simplifier for RHF by
        obtaining the exchange hole directly as K Matrix without doing
        double integral by quadrature.
        """
        if self.vxc_hole_WF is not None and grid_info is None:
            return self.vxc_hole_WF

        if self.eng.wfn.name() == "CIWavefunction":
            Tau_ijkl = self.eng.wfn.get_tpdm("SUM", True).np
            D2 = self.eng.wfn.get_opdm(-1, -1, "SUM", True).np
            C = self.eng.wfn.Ca()
            Tau_ijkl = contract("pqrs,ip,jq,ur,vs->ijuv", Tau_ijkl, C, C, C, C)
            D2 = C.np @ D2 @ C.np.T
        else:
            D2a = self.eng.wfn.Da().np
            D2b = self.eng.wfn.Db().np
            D2 = D2a + D2b

        if grid_info is None:
            vxchole = np.zeros(Vpot.grid().npoints())
            nblocks = Vpot.nblocks()
            points_func_outer = Vpot.properties()[0]
            blocks = None
        else:
            blocks, npoints, points_func_outer = grid_info
            vxchole = np.zeros(npoints)
            nblocks = len(blocks)
            points_func_outer.set_deriv(0)

        points_func = Vpot.properties()[0]
        points_func.set_deriv(0)

        # First loop over the outer set of blocks
        num_block_ten_percent = int(nblocks / 10)
        w1_old = 0
        print(f"vxchole quadrature double integral starts ({(Vpot.grid().npoints()):d} points): ", end="")
        start_time = time.time()
        for l_block in range(nblocks):
            # Print out progress
            if num_block_ten_percent != 0 and l_block % num_block_ten_percent == 0:
                print(".", end="")
    
            # Obtain general grid information
            if blocks is None:
                l_grid = Vpot.get_block(l_block)
            else:
                l_grid = blocks[l_block]

            l_x = np.array(l_grid.x())
            l_y = np.array(l_grid.y())
            l_z = np.array(l_grid.z())
            l_npoints = l_x.shape[0]

            points_func_outer.compute_points(l_grid)

            l_lpos = np.array(l_grid.functions_local_to_global())
            l_phi = np.array(points_func_outer.basis_values()["PHI"])[:l_npoints, :l_lpos.shape[0]]
    
            # if restricted:
            lD1 = D2[(l_lpos[:, None], l_lpos)]
            rho1 = contract('pm,mn,pn->p', l_phi, lD1, l_phi)
            rho1inv = (1 / rho1)[:, None]

            dvp_l = np.zeros_like(l_x)

            # if not restricted:
            #     dvp_l_b = np.zeros_like(l_x)
    
            # Loop over the inner set of blocks
            for r_block in range(Vpot.nblocks()):
                r_grid = Vpot.get_block(r_block)
                r_w = np.array(r_grid.w())
                r_x = np.array(r_grid.x())
                r_y = np.array(r_grid.y())
                r_z = np.array(r_grid.z())
                r_npoints = r_w.shape[0]
    
                points_func.compute_points(r_grid)
    
                r_lpos = np.array(r_grid.functions_local_to_global())
    
                # Compute phi!
                r_phi = np.array(points_func.basis_values()["PHI"])[:r_npoints, :r_lpos.shape[0]]
    
                # Build a local slice of D
                if self.eng.wfn.name() == "CIWavefunction":
                    lD2 = D2[(r_lpos[:, None], r_lpos)]
                    rho2 = contract('pm,mn,pn->p', r_phi, lD2, r_phi)
    
                    p, q, r, s = np.meshgrid(l_lpos, l_lpos, r_lpos, r_lpos, indexing="ij")
                    Tap_temp = Tau_ijkl[p, q, r, s]
    
                    n_xc = contract("mnuv,pm,pn,qu,qv->pq", Tap_temp, l_phi, l_phi, r_phi, r_phi)
                    n_xc *= rho1inv
                    n_xc -= rho2
                elif self.eng.wfn.name() == "RHF":
                    lD2 = self.eng.wfn.Da().np[(l_lpos[:, None], r_lpos)]
                    n_xc = - 2 * contract("mu,nv,pm,pn,qu,qv->pq", lD2, lD2, l_phi, l_phi, r_phi, r_phi)
                    n_xc *= rho1inv

                # Build the distnace matrix
                R2 = (l_x[:, None] - r_x) ** 2
                R2 += (l_y[:, None] - r_y) ** 2
                R2 += (l_z[:, None] - r_z) ** 2
                # R2 += 1e-34
                mask1 = R2 <= atol
                mask2 = (R2 > atol) * (R2 < atol1)
                if np.any(mask1 + mask2):
                    R2[mask1] = np.inf
                    R2[mask2] = atol1
                Rinv = 1 / np.sqrt(R2)
    
                # if restricted:
                dvp_l += np.sum(n_xc * Rinv * r_w, axis=1)

            # if restricted:
            vxchole[w1_old:w1_old + l_npoints] += dvp_l
            w1_old += l_npoints
    
        print("\n")
        print(f"Totally {vxchole.shape[0]} grid points takes {(time.time() - start_time):.2f}s "
              f"with max {psi4.core.get_global_option('DFT_BLOCK_MAX_POINTS')} points in a block.")
        assert w1_old == vxchole.shape[0], "Somehow the whole space is not fully integrated."
        if blocks is None:
            # if restricted:
            self.vxc_hole_WF = vxchole
            return self.vxc_hole_WF
        else:
            # if restricted:
            return vxchole

    def mRKS(self, maxiter, vxc_grid=None, v_tol=1e-4, D_tol=1e-7,
             eig_tol=1e-4, frac_old=0.5, init="scan",
             sing=(1e-5, 1e-4, 1e-5, 1e-4)):
        """
        the modified Ryabinkin-Kohut-Staroverov method.

        Currently it supports two different kind of input wavefunction:
            1) Psi4.CIWavefunction
            2) Psi4.RHF
        and it only supports spin-restricted wavefunction.
        Side note: spin-unrestricted HF wavefunction (psi4.UHF) can easily
        be supported but unrestricted CI or restricted/unrestricted CCSD
        can not, because of the absence of tpdm in both methods.

        parameters:
        ----------------------
            maxiter: int
                same as opt_max_iter
            vxc_grid: np.ndarray of shape (3, num_grid_points), opt
                When this is given, the final result will be represented
            v_tol: float, opt
                convergence criteria for vxc Fock matrices.
                default: 1e-4
            D_tol: float, opt
                convergence criteria for density matrices.
                default: 1e-7
            eig_tol: float, opt
                convergence criteria for occupied eigenvalue spectrum.
                default: 1e-4
            frac_old: float, opt
                Linear mixing parameter for current vxc and old vxc.
                If 0, no old vxc is mixed in.
                Should be in [0,1)
                default: 0.5.
            init: string, opt
                Initial guess method.
                default: "SCAN"
                1) If None, input wfn info will be used as initial guess.
                2) If "continue" is given, then it will not initialize
                but use the densities and orbitals stored. Meaningly,
                one can run a quick WY calculation as the initial
                guess. This can also be used to user speficified
                initial guess by setting Da, Coca, eigvec_a.
                3) If it's not continue, it would be expecting a
                method name string that works for psi4. A separate psi4 calculation
                would be performed.
            sing: tuple of float of length 4, opt.
                Singularity parameter for _vxc_hole_quadrature()
                default: (1e-5, 1e-4, 1e-5, 1e-4)
                [0]: atol, [1]: atol1 for dft_spherical grid calculation.
                [2]: atol, [3]: atol1 for vxc_grid calculation.


        returns:
        ----------------------
            The result will be save as self.grid.vxc
    """
        if not self.eng.wfn.name() in ["CIWavefunction", "RHF"]:
            raise ValueError("Currently only supports Psi4 CI wavefunction"
                             "inputs because Psi4 CCSD wavefunction currently "
                             "does not support two-particle density matrices.")

        if self.ref != 1:
            raise ValueError("Currently only supports Spin-Restricted "
                             "calculations since Spin-Unrestricted CI "
                             "is not supported by Psi4.")

        if self.guide_components != "hartree":
            raise ValueError("Hartree potential is necessary as the guide potential.")

        Nalpha = self.nalpha

        # Preparing DFT spherical grid
        functional = psi4.driver.dft.build_superfunctional("SVWN", restricted=True)[0]
        Vpot = psi4.core.VBase.build(self.eng.basis, functional, "RV")
        Vpot.initialize()
        Vpot.properties()[0].set_pointers(self.eng.wfn.Da())


        # Preparing for WF properties
        if self.eng.wfn.name() == "CIWavefunction":
            if not (psi4.core.get_global_option("opdm") and psi4.core.get_global_option("tpdm")):
                raise ValueError("For CIWavefunction as input, make sure to turn on opdm and tpdm.")
            # TPDM & ERI Memory check
            nbf = self.nbf
            I_size = (nbf ** 4) * 8.e-9 * 2
            numpy_memory = 2
            memory_footprint = I_size * 1.5
            if I_size > numpy_memory:
                psi4.core.clean()
                raise Exception("Estimated memory utilization (%4.2f GB) exceeds allotted memory \
                                            limit of %4.2f GB." % (memory_footprint, numpy_memory))
            else:
                print("Memory taken by ERI integral matrix and 2pdm is about: %.3f GB." % memory_footprint)


            opdm = np.array(self.eng.wfn.get_opdm(-1,-1,"SUM",False))
            tpdm = self.eng.wfn.get_tpdm("SUM", True).np

            Ca = self.eng.wfn.Ca().np

            mints = psi4.core.MintsHelper(self.eng.basis)

            Ca_psi4 = psi4.core.Matrix.from_array(Ca)
            I = mints.mo_eri(Ca_psi4, Ca_psi4, Ca_psi4, Ca_psi4).np
            # I = 0.5 * I + 0.25 * np.transpose(I, [0, 1, 3, 2]) + 0.25 * np.transpose(I, [1, 0, 2, 3])
            # Transfer the AO h into MO h
            h = Ca.T @ (self.T + self.V) @ Ca

            # Generalized Fock Matrix is constructed on the
            # basis of MOs, which are orthonormal.
            F_GFM = opdm @ h + contract("rsnq,rsmq->mn", I, tpdm)
            F_GFM = 0.5 * (F_GFM + F_GFM.T)

            del mints, I

            C_a_GFM = psi4.core.Matrix(nbf, nbf)
            eigs_a_GFM = psi4.core.Vector(nbf)
            psi4.core.Matrix.from_array(F_GFM).diagonalize(C_a_GFM,
                                                           eigs_a_GFM,
                                                           psi4.core.DiagonalizeOrder.Ascending)

            eigs_a_GFM = eigs_a_GFM.np / 2.0  # RHF
            C_a_GFM = C_a_GFM.np
            # Transfer to AOs
            C_a_GFM = Ca @ C_a_GFM

            # Solving for Natural Orbitals (NO) 
            C_a_NO = psi4.core.Matrix(nbf, nbf)
            eigs_a_NO = psi4.core.Vector(nbf)
            psi4.core.Matrix.from_array(opdm).diagonalize(C_a_NO, eigs_a_NO, psi4.core.DiagonalizeOrder.Descending)
            eigs_a_NO = eigs_a_NO.np / 2.0  # RHF
            C_a_NO = C_a_NO.np
            C_a_NO = Ca @ C_a_NO

            # Rrepare components on the grid
            ebarWF      = self.eng.grid._average_local_orbital_energy(self.Dt[0], C_a_GFM, eigs_a_GFM, Vpot=Vpot)
            taup_rho_WF = self.eng.grid._pauli_kinetic_energy_density(self.Dt[0], C_a_NO, eigs_a_NO, Vpot=Vpot)
        elif self.eng.wfn.name() == "RHF":  # Since HF is a special case, there is no need for GFM and NO as in CI.
            epsilon_a = self.eng.wfn.epsilon_a_subset("AO", "OCC").np
            ebarWF = self.eng.grid._average_local_orbital_energy(self.Dt[0], self.ct[0][:,:Nalpha], epsilon_a, Vpot=Vpot)
            taup_rho_WF = self.eng.grid._pauli_kinetic_energy_density(self.Dt[0], self.ct[0], Vpot=Vpot)
        else:
            raise ValueError("Currently only supports Spin-Restricted "
                             "calculations since Spin-Unrestricted CI "
                             "is not supported by Psi4.")

        vxchole = self._vxc_hole_quadrature(atol=sing[0], atol1=sing[1], Vpot=Vpot)

        emax = np.max(ebarWF)

        # Initialization.
        if init is None:
            self.Da = np.copy(self.Dt[0])
            self.Coca = np.copy(self.ct[0])
            self.eigvecs_a = self.eng.wfn.epsilon_a().np[:Nalpha]
        elif init.lower()=="continue":
            pass
        else:
            wfn_temp = psi4.energy(init+"/" + self.eng.basis_str, molecule=self.eng.mol, return_wfn=True)[1]
            self.Da = np.array(wfn_temp.Da())
            self.Coca = np.array(wfn_temp.Ca())[:, :Nalpha]
            self.eigvecs_a = np.array(wfn_temp.epsilon_a())
            del wfn_temp

        vxc_old = 0.0
        Da_old = 0.0
        eig_old = 0.0
        for mRKS_step in range(1, maxiter+1):
            # ebarKS = self._average_local_orbital_energy(self.molecule.Da.np, self.molecule.Ca.np[:,:Nalpha], self.molecule.eig_a.np[:Nalpha] + self.vout_constant)
            ebarKS      = self.eng.grid._average_local_orbital_energy(self.Da, self.Coca, self.eigvecs_a[:Nalpha], Vpot=Vpot)
            taup_rho_KS = self.eng.grid._pauli_kinetic_energy_density(self.Da, self.Coca, Vpot=Vpot)
            # self.vout_constant = emax - self.molecule.eig_a.np[self.molecule.nalpha - 1]
            potential_shift = emax - np.max(ebarKS)
            self.shift = potential_shift

            vxc = vxchole + ebarKS - ebarWF + taup_rho_WF - taup_rho_KS + potential_shift

            # Add compulsory mixing parameter close to the convergence to help convergence HOPEFULLY

            verror = np.linalg.norm(vxc - vxc_old) / self.nbf ** 2
            if verror < v_tol:
                print("vxc stops updating.")
                break

            Derror = np.linalg.norm(self.Da - Da_old) / self.nbf ** 2
            eerror = (np.linalg.norm(self.eigvecs_a[:Nalpha] - eig_old) / Nalpha)
            if (Derror < D_tol) and (eerror < eig_tol):
                print("KSDFT stops updating.")
                break

            # linear Mixture
            if mRKS_step != 1:
                vxc = vxc * (1 - frac_old) + vxc_old * frac_old

            # Save old data.
            vxc_old = np.copy(vxc)
            Da_old = np.copy(self.Da)
            eig_old = np.copy(self.eigvecs_a[:Nalpha])

            vxc_Fock = self.eng.grid.dft_grid_to_fock(vxc, Vpot)

            self._diagonalize_with_potential_vFock(v=vxc_Fock)

            print(f"Iter: {mRKS_step}, Density Change: {Derror:2.2e}, Eigenvalue Change: {eerror:2.2e}, "
                  f"Potential Change: {verror:2.2e}.")

        if vxc_grid is not None:
            grid_info = self.eng.grid.grid_to_blocks(vxc_grid)
            grid_info[-1].set_pointers(self.eng.wfn.Da())

            # A larger atol seems to be necessary for user-defined grid
            vxchole = self._vxc_hole_quadrature(grid_info=grid_info,
                                                atol=sing[2], atol1=sing[3], Vpot=Vpot)
            if self.eng.wfn.name() == "CIWavefunction":
                ebarWF = self.eng.grid._average_local_orbital_energy(self.Dt[0],
                                                            C_a_GFM, eigs_a_GFM,
                                                            grid_info=grid_info,Vpot=Vpot)
                taup_rho_WF = self.eng.grid._pauli_kinetic_energy_density(self.Dt[0],
                                                                 C_a_NO, eigs_a_NO,
                                                                 grid_info=grid_info, Vpot=Vpot)
            elif self.eng.wfn.name() == "RHF":
                ebarWF = self.eng.grid._average_local_orbital_energy(self.Dt[0],
                                                            self.ct[0],
                                                            epsilon_a[:Nalpha],
                                                            grid_info=grid_info, Vpot=Vpot)
                taup_rho_WF = self.eng.grid._pauli_kinetic_energy_density(self.Dt[0],
                                                                 self.ct[0],
                                                                 grid_info=grid_info, Vpot=Vpot)
            ebarKS = self.eng.grid._average_local_orbital_energy(self.Da, self.Coca,
                                                        self.eigvecs_a[:Nalpha], grid_info=grid_info)
            taup_rho_KS = self.eng.grid._pauli_kinetic_energy_density(self.Da, self.Coca,
                                                             grid_info=grid_info, Vpot=Vpot)

            potential_shift = np.max(ebarWF) - np.max(ebarKS)
            self.shift = potential_shift

            vxc = vxchole + ebarKS - ebarWF + taup_rho_WF - taup_rho_KS + potential_shift
        self.grid_vxc = vxc
        return

    def _diagonalize_with_potential_vFock(self, v=None):
        """
        Diagonalize Fock matrix with additional external potential
        """

        if v is None:
            fock_a = self.V + self.T + self.va
        else:
            if self.ref == 1:
                fock_a = self.V + self.T + self.va + v
            else:
                valpha, vbeta = v
                fock_a = self.V + self.T + self.va + valpha
                fock_b = self.V + self.T + self.vb + vbeta


        self.Ca, self.Coca, self.Da, self.eigvecs_a = self.diagonalize( fock_a, self.nalpha )

        if self.ref == 1:
            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.Ca.copy(), self.Coca.copy(), self.Da.copy(), self.eigvecs_a.copy()
        else:
            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.diagonalize( fock_b, self.nbeta )


try:
    import pyscf
    has_pyscf = True
except ImportError:
    has_pyscf = False

try:
    import psi4
    has_psi4 = True
except ImportError:
    has_psi4 = False

if has_pyscf:
    from .pyscfgrider import PySCFGrider
if has_psi4:
    from .psi4grider import Psi4Grider



"""
engines.py
"""

from abc import ABC, abstractmethod

class Engine(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def set_system(self, molecule, basis, ref, pbs):
        """
        Set system into engine. It is expected in this function to format each
        component in the appropriate way for the engine. 
        The inverter will require as well the following data:
        mol_str: str
            Molecule information in string form
        nbf: int
            Number of basis functions
        npbs: int
            Number of basis functions in basis used to express the inverted
            potential
        """
        pass

    @abstractmethod
    def initialize(self):
        """
        Initializes different components needed for inversions. This can include
        the generation of the overlap and electric repulsion matrices. As well
        as setting Coulomb/Exchange objects
        """
        pass

    @abstractmethod
    def get_V(self):
        """
        Generates nuclear external potential on the atomic orbital basis set. 
        
        Parameters:
        ----------
        
        Returns:
        --------
        V: np.ndarray
            External potential on basis. Size: (nbf, nbf)
        """
        return V

    @abstractmethod
    def get_T(self):
        """
        Generates Kinetic contribution on the atomic orbital basis set. 
        
        Parameters:
        ----------
        
        Returns:
        --------
        T: np.ndarray
            Kinetic matrix on basis. Size: (nbf, nbf)
        """
        return T

    @abstractmethod
    def get_Tpbas(self):
        """
        Generates Kinetic contribution on the atomic orbital basis set used
        to express the inverted potential
        
        Parameters:
        ----------
        
        Returns:
        --------
        Tpbas: np.ndarray
            Kinetic matrix on secondary basis. Size: (npbas, npbas)
        """
        return Tpbas 

    @abstractmethod
    def get_A(self):
        """
        Generates nuclear external potential on the atomic orbital basis set. 
        
        Parameters:
        ----------
        
        Returns:
        --------
        V: np.ndarray
            External potential on basis. Size: (nbf, nbf)
        """
        return A

    @abstractmethod
    def get_S(self):
        """
        Generates the overlap matrix of the atomic orbital basis set. 
        
        Parameters
        ----------
        
        Returns:
        --------
        S: np.ndarray
            Overlap matrix. Size: (nbf, nbf)
        """
        return S

    @abstractmethod
    def get_S3(self):
        """
        Generates the three overlap matrix of the atomic orbital basis set with
        the basis used to express the inverted potential. 
        
        Parameters:
        ----------
        
        Returns:
        --------
        S3: np.ndarray
            Three overlap matrix. Size: (nbf, nbf, npbas)
        """
        return S3

    @abstractmethod
    def get_S4(self):
        """
        Generates the four overlap matrix of 4 different orbital basis sets. 
        
        Parameters:
        ----------
        
        Returns:
        --------
        S4: np.ndarray
            Four overlap matrix
        """
        return S4

    @abstractmethod
    def compute_hartree(self, Cocc_a, Cocc_b):
        """
        Generates the Hartree potential on the atomic orbital basis set
        
        Parameters:
        -----------
        Cocc_a, Cocc,b: np.ndarray. Size: (nbf, nbf)
            Occupied molecular orbitals (alpha, beta)

        Returns:
        --------
        J: List of lenght two with np.ndarrays. 
            Hartree potential on ao basis (alpha, beta)
        """
        return J
    






        


"""
Inverter.py
"""
from warnings import warn
from dataclasses import dataclass
import numpy as np
from opt_einsum import contract

from .methods.zmp import ZMP
from .methods.wuyang import WuYang
from .methods.pdeco import PDECO
from .methods.oucarter import OC
from .methods.mrks import MRKS
from .methods.direct import Direct

@dataclass
class V:
    """Stores Potentials on AO"""
    T : np.ndarray

class E:
    """Stores Energies"""

class Inverter(Direct, ZMP, WuYang, PDECO, OC, MRKS):
    """
    Attributes:
    ----------

    mol : Engine.molecule
        Molecule class of engine used
    basis : Engine.basis
        Basis class of engine used
    basis_str : str
        Basis set
    nbf : int
        Number of basis functions for main calculation
    nalpha : int
        Number of alpha electrons
    nbeta : int
        Number of beta electrons
    ref : {1,2}
        Reference calculation
        1 -> Restricted
        2 -> Unrestricted
    Dt : List
        List of np.ndarray for target density matrices (on AO).
    ct : List
        List of np.ndarray for input occupied orbitals. This might not be correct for post-HartreeFock methods.
    pbs_str: string
        name of Potential basis set
    pbs : Engine.basis
        Basis class for Potential basis set of the engine used. 
    npbs : int
        the length of pbs
    v_pbs : np.ndarray shape (npbs, ) for ref==1 and (2*npbs, ) for ref==2.
        potential vector on the Potential Baiss Set.
        If the potential is not represented on the basis set, this should
        remain 0. It will be initialized to a 0 array. One can set this
        value for initial guesses before Wu-Yang method (WY) or PDE-Constrained
        Optimization method (PDE-CO). For example, if PDE-CO is ran after
        a WY calculation, the initial for PDE-CO will be the result of WY
        if v_pbs is not zeroed.
    S2  : np.ndarray
        The ao overlap matrix (i.e. S matrix)
    S3  : np.ndarray
        The three ao overlap matrix (ao, ao, pbs)
    S4  : np.ndarray
        The four ao overlap matrix, the size should be (ao, ao, ao, ao)
    jk  : Engine.jk
        Engine jk object.
    T   : np.ndarray
        kinetic matrix on ao
    V   : np.ndarray
        external potential matrix on ao
    T_pbs: np.ndarray
        kinetic matrix on pbs. Useful for regularization.
    guide_potential_components: list of string
        guide potential components name
    va, vb: np.ndarray of shape (nbasis, nbasis)
        guide potential Fock matrix.
    """

    def __init__( self, engine='psi4' ):
        self.eng_str = engine.lower()
        if engine.lower() == 'psi4':
            from .engines import Psi4Engine
            self.eng = Psi4Engine()
        elif engine.lower() == 'pyscf':
            from .engines import PySCFEngine
            self.eng = PySCFEngine()
        else:
            raise ValueError("Engine name is incorrect. The availiable engines are: {psi4, pyscf}")
            
    def __repr__( self ):
        return "n2v.Inverter"

    def set_system( self, molecule, basis, ref=1, pbs='same' , **kwargs):
        """
        Stores relevant information and intitializes Engine

        Parameters
        ----------
        molecule: Engine.molecule
            Molecule object of selected engine
        basis: str
            Basis set of the main calculation
        ref: int
            reference for system. Restricted   -> 1
                                  Unrestricted -> 2
        pbs: str, default='same'
            Basis set for the potential
        **kwargs:
            Optional Parameters for different Engiens 
            Psi4 Engine:
                wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
                Psi4 wavefunction object
            PySCF Engine:
                None

        """
        # Communicate TO engine

        self.eng.set_system(molecule, basis, ref, pbs, **kwargs)
        self.ref = ref

        self.nalpha = self.eng.nalpha
        self.nbeta = self.eng.nbeta

        # Initialize ecompasses everything the engine builds with basis set 
        self.eng.initialize()
        self.set_basis_matrices()

        # Receive FROM engine
        self.nbf  = self.eng.nbf
        self.npbs = self.eng.npbs
        self.v_pbs = np.zeros( (self.npbs) ) if self.ref == 1 \
                                             else np.zeros( 2 * self.npbs )

    @classmethod
    def from_wfn( self, wfn, pbs='same' ):
        """
        Generates Inverter directly from wavefunction. 
        
        Parameters
        ----------
        wfn: Psi4.Core.{RHF, RKS, ROHF, CCWavefunction, UHF, UKS, CUHF}
            Wavefunction Object
        Returns
        -------
        inv: n2v.Inverter
            Inverter Object. 
        """
        from .engines import Psi4Engine
        inv = self( engine='psi4' )
        inv.eng = Psi4Engine()
        ref = 1 if wfn.to_file()['boolean']['same_a_b_dens'] else 2
        inv.set_system( wfn.molecule(), wfn.basisset().name(), pbs=pbs, ref=ref, wfn=wfn )
        inv.Dt = [ np.array(wfn.Da()), np.array(wfn.Db()) ]
        inv.ct = [ np.array(wfn.Ca_subset("AO", "OCC")), np.array(wfn.Cb_subset("AO", "OCC")) ]
        inv.et = [ np.array(wfn.epsilon_a_subset("AO", "OCC")), np.array(wfn.epsilon_b_subset("AO", "OCC")) ]
        inv.eng_str = 'psi4'
        inv.eng.wfn = wfn

        return inv

    def set_basis_matrices( self ):
        """
        Generate basis dependant matrices
        """
        self.T  = self.eng.get_T()
        self.V  = self.eng.get_V()
        self.A  = self.eng.get_A()
        self.S2 = self.eng.get_S()
        self.S3      = self.eng.get_S3()

        if self.eng.pbs_str != 'same':  
            self.T_pbs  = self.eng.get_Tpbas()

        self.S4 = None

    def compute_hartree( self, Cocc_a, Cocc_b ):
        """
        Computes Hartree Potential on AO basis set. 

        Parameters
        ----------
        Cocc_a, Cocc_b: np.ndarray (nbf, nbf)
            Occupied orbitals in ao basis

        Returns
        -------
        J: List of np.ndarray
            Hartree potential due to density from Cocc_a and Cocc_b
        """
        return self.eng.compute_hartree(Cocc_a, Cocc_b )

    def diagonalize( self, matrix, ndocc ):
        """
        Diagonalizes Fock Matrix

        Parameters
        ----------
        marrix: np.ndarray
            Matrix to be diagonalized
        ndocc: int
            Number of occupied orbitals

        Returns
        -------
        C: np.ndarray
            Orbital Matrix
        Cocc: np.ndarray
            Occupied Orbital Matrix
        D: np.ndarray
            Density Matrix
        eigves: np.ndarray
            Eigenvalues
        """

        Fp = self.A.dot(matrix).dot(self.A)
        eigvecs, Cp = np.linalg.eigh(Fp)
        C = self.A.dot(Cp)
        Cocc = C[:, :ndocc]
        D = contract('pi,qi->pq', Cocc, Cocc)
        return C, Cocc, D, eigvecs

    def diagonalize_with_potential_vFock(self, v=None):
        """
        Diagonalize Fock matrix with additional external potential
        Stores values in object. 

        Parameters
        ----------
        v: np.ndarray
            Additional external potential to be added to hamiltonian along with:
            Kinetic_nm
            External_nm
            Guide_Potential_nm
        """

        if v is None:
            fock_a = self.V + self.T + self.va
        else:
            if self.ref == 1:
                fock_a = self.V + self.T + self.va + v
            else:
                valpha, vbeta = v
                fock_a = self.V + self.T + self.va + valpha
                fock_b = self.V + self.T + self.vb + vbeta


        self.Ca, self.Coca, self.Da, self.eigvecs_a = self.diagonalize( fock_a, self.nalpha )

        if self.ref == 1:
            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.Ca.copy(), self.Coca.copy(), self.Da.copy(), self.eigvecs_a.copy()
        else:
            self.Cb, self.Cocb, self.Db, self.eigvecs_b = self.diagonalize( fock_b, self.nbeta )    

    # Actual Methods
    def generate_components(self, guide_components, **keywords):
        """
        Generates exact potential components to be added to
        the Hamiltonian to aide in the inversion procedure. 
        Parameters:
        -----------
        guide_potential_components: list
            Components added as to guide inversion. 
            Can be chosen from ["hartree", "fermi_amandi", "svwn"]
        """

        self.guide_components = guide_components
        self.va = np.zeros( (self.nbf, self.nbf) )
        self.vb = np.zeros( (self.nbf, self.nbf) )
        self.J0 = self.compute_hartree(self.ct[0], self.ct[1])
        N       = self.nalpha + self.nbeta

        if self.eng_str == 'psi4':
            J0_NO = self.eng.hartree_NO(self.Dt[0])
            self.J0 = J0_NO if J0_NO is not None else self.J0

        if guide_components == 'none':
            warn("No guide potential was provided. Convergence may not be achieved")
        elif guide_components == 'hartree':
            self.va += self.J0[0] + self.J0[1]
            self.vb += self.J0[0] + self.J0[1]
        elif guide_components == 'fermi_amaldi':
            v_fa = (1-1/N) * (self.J0[0] + self.J0[1])
            self.va += v_fa
            self.vb += v_fa
        else:
            raise ValueError("Guide component not recognized")

    def invert(self, method, 
                     guide_components = 'hartree',
                     opt_max_iter = 50,
                     **keywords):
        """
        Handler to all available inversion methods
        Parameters
        ----------
        method: str
            Method used to invert density. 
            Can be chosen from {wuyang, zmp, mrks, oc}. 
            See documentation below for each method. 
        guide_components: list, opt
            Components added as to guide inversion. 
            Can be chosen from {"fermi_amandi", "svwn"}
            Default: ["fermi_amaldi"]
        opt_max_iter: int, opt
            Maximum number of iterations inside the chosen inversion.
            Default: 50
        direct
        ------
        Direct inversion of a set of Kohn-Sham equations. 
        $$v_{xc}(r) = \frac{1}{n(r)} \sum_i^N [\phi_i^{*} (r) \nabla^2 \phi_i(r) + \varepsilon_i | \phi_i(r)|^2] $$
            Parameters:
            -----------
                grid: np.ndarray, opt
                    Grid where result will be expressed in.
                    If not provided, dft grid will be used instead. 
                
        wuyang
        ------
        the Wu-Yang method:
        The Journal of chemical physics 118.6 (2003): 2498-2509.
            Parameters:
            ----------
                opt_max_iter: int
                    maximum iteration
                opt_method: string, opt
                    Method for scipy optimizer
                    Currently only used by wuyang and pdeco method.
                    Defaul: 'trust-krylov'
                    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
                reg : float, opt
                    Regularization constant for Wuyant Inversion.
                    Default: None -> No regularization is added.
                    Becomes attribute of inverter -> inverter.lambda_reg
                tol: float
                    tol for scipy.optimize.minimize
                gtol: float
                    gtol for scipy.optimize.minimize: the gradient norm for
                    convergence
                opt: dict
                    options for scipy.optimize.minimize
                                Notice that opt has lower priorities than opt_max_iter and gtol.
            return:
                the result are stored in self.v_pbs
            
        zmp
        ---
        The Zhao-Morrison-Parr Method:
        Phys. Rev. A 50, 2138 
            Parameters:
            ----------
                lambda_list: list
                    List of Lamda parameters used as a coefficient for Hartree 
                    difference in SCF cycle.
                zmp_mixing: float, optional
                    mixing \in [0,1]. How much of the new potential is added in.
                    For example, zmp_mixing = 0 means the traditional ZMP, i.e. all the potentials from previous
                    smaller lambda are ignored.
                    Zmp_mixing = 1 means that all the potentials of previous lambdas are accumulated, the larger lambda
                    potential are meant to fix the wrong/inaccurate region of the potential of the sum of the previous
                    potentials instead of providing an entire new potentials.
                    default: 1
                opt_max_iter: float
                    Maximum number of iterations for scf cycle
                opt_tol: float
                    Convergence criteria set for Density Difference and DIIS error. 
                return:
                    The result will be stored in self.proto_density_a and self.proto_density_b
                    For zmp_mixing==1, restricted (ref==1):
                        self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0])
                        self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[1]);
                    unrestricted (ref==1):
                        self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0] + Dt[1])
                        self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[0] + Dt[1]);
                    For restricted (ref==1):
                        vxc = \int dr' \frac{self.proto_density_a + self.proto_density_b}{|r-r'|}
                            = 2 * \int dr' \frac{self.proto_density_a}{|r-r'|};
                    for unrestricted (ref==2):
                        vxc_up = \int dr' \frac{self.proto_density_a}{|r-r'|}
                        vxc_down = \int dr' \frac{self.proto_density_b}{|r-r'|}.
                    To get potential on grid, one needs to do
                        vxc = self.on_grid_esp(Da=self.proto_density_a, Db=self.proto_density_b, grid=grid) for restricted;
                        vxc_up = self.on_grid_esp(Da=self.proto_density_a, Db=np.zeros_like(self.proto_density_a),
                                  grid=grid) for unrestricted;
        mRKS
        ----
        the modified Ryabinkin-Kohut-Staroverov method:
        Phys. Rev. Lett. 115, 083001 
        J. Chem. Phys. 146, 084103p
            Parameters:
            -----------
                maxiter: int
                    same as opt_max_iter
                vxc_grid: np.ndarray of shape (3, num_grid_points), opt
                    When this is given, the final result will be represented
                v_tol: float, opt
                    convergence criteria for vxc Fock matrices.
                    default: 1e-4
                D_tol: float, opt
                    convergence criteria for density matrices.
                    default: 1e-7
                eig_tol: float, opt
                    convergence criteria for occupied eigenvalue spectrum.
                    default: 1e-4
                frac_old: float, opt
                    Linear mixing parameter for current vxc and old vxc.
                    If 0, no old vxc is mixed in.
                    Should be in [0,1)
                    default: 0.5.
                init: string or psi4.core.Wavefunction, opt
                    Initial guess method.
                    default: "SCAN"
                    1) If None, input wfn info will be used as initial guess.
                    2) If "continue" is given, then it will not initialize
                    but use the densities and orbitals stored. Meaningly,
                    one can run a quick WY calculation as the initial
                    guess. This can also be used to user speficified
                    initial guess by setting Da, Coca, eigvec_a.
                    3) If it's not continue, it would be expecting a
                    method name string that works for psi4. A separate psi4 calculation
                    would be performed.
                sing: tuple of float of length 4, opt.
                    Singularity parameter for _vxc_hole_quadrature()
                    default: (1e-5, 1e-4, 1e-5, 1e-4)
                    [0]: atol, [1]: atol1 for dft_spherical grid calculation.
                    [2]: atol, [3]: atol1 for vxc_grid calculation.
            return:
                The result will be stored in self.grid.vxc
        oc
        --
        Ou-Carter method
        J. Chem. Theory Comput. 2018, 14, 5680−5689
            Parameters:
            -----------
                maxiter: int
                    same as opt_max_iter
                vxc_grid: np.ndarray of shape (3, num_grid_points)
                    The final result will be represented on this grid
                    default: 1e-4
                D_tol: float, opt
                    convergence criteria for density matrices.
                    default: 1e-7
                eig_tol: float, opt
                    convergence criteria for occupied eigenvalue spectrum.
                    default: 1e-4
                frac_old: float, opt
                    Linear mixing parameter for current vxc and old vxc.
                    If 0, no old vxc is mixed in.
                    Should be in [0,1)
                    default: 0.5.
                init: string, opt
                    Initial guess method.
                    default: "SCAN"
                    1) If None, input wfn info will be used as initial guess.
                    2) If "continue" is given, then it will not initialize
                    but use the densities and orbitals stored. Meaningly,
                    one can run a quick WY calculation as the initial
                    guess. This can also be used to user speficified
                    initial guess by setting Da, Coca, eigvec_a.
                    3) If it's not continue, it would be expecting a
                    method name string that works for psi4. A separate psi4 calculation
                    would be performed.
                    wuyang
        pdeco
        -----
        the PDE-Constrained Optimization method:
        Int J Quantum Chem. 2018;118:e25425;
        Nat Commun 10, 4497 (2019).
            Parameters:
            ----------
                opt_max_iter: int
                    maximum iteration
                opt_method: string, opt
                    Method for scipy optimizer
                    Currently only used by wuyang and pdeco method.
                    Defaul: 'L-BFGS-B'
                    Options: ['L-BFGS-B', 'BFGS']
                    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
                reg : float, opt
                    Regularization constant for Wuyant Inversion.
                    Default: None -> No regularization is added.
                    Becomes attribute of inverter -> inverter.lambda_reg
                gtol: float
                    gtol for scipy.optimize.minimize: the gradient norm for
                    convergence
                opt: dict
                    options for scipy.optimize.minimize
                    Notice that opt has lower priorities than opt_max_iter and gtol.
            return:
                the result are stored in self.v_pbs
        """

        self.generate_components(guide_components)

        if method.lower() == "direct":
            return self.direct_inversion(**keywords)
        elif method.lower() == "wuyang":
            self.wuyang(opt_max_iter, **keywords)
        elif method.lower() == "zmp":
            self.zmp(opt_max_iter, **keywords)
        elif method.lower() == "mrks":
            if self.eng_str == 'pyscf':
                raise ValueError("mRKS method not yet available with the PySCF engine. Try another method or another engine.")
            return self.mRKS(opt_max_iter, **keywords)
        elif method.lower() == 'oc':
            if self.eng_str == 'pyscf':
                raise ValueError("OuCarter method not yet available with the PySCF engine. Try another method or another engine.")
            return self.oucarter(opt_max_iter, **keywords)
        elif method.lower() == 'pdeco':
            return self.pdeco(opt_max_iter, **keywords)
        else:
            raise ValueError(f"Inversion method not available. Methods available: {['wuyang', 'zmp', 'mrks', 'oc', 'pdeco']}")




"""
grider_pyscf.py
Grider for PyScf
"""

from gbasis.evals.density import (evaluate_density, 
#                                  evaluate_posdef_kinetic_energy_density,
                                  evaluate_density_laplacian,
                                  evaluate_density_gradient,
 #                                 evaluate_general_kinetic_energy_density,
                                )
from gbasis.evals.eval import evaluate_basis
from gbasis.evals.eval_deriv import evaluate_deriv_basis
from gbasis.evals.electrostatic_potential import point_charge_integral
from gbasis.wrappers import from_pyscf

# from grid.molgrid import MolGrid
# from grid.onedgrid import GaussLaguerre,  GaussLegendre, HortonLinear
# from grid.becke import BeckeWeights

import numpy as np
from opt_einsum import contract

try:
    from pyscf import dft
    has_pyscf = True
except ImportError:
    has_pyscf = False

if has_pyscf:
    class PySCFGrider:
        """
        PySCF Grider Class
        Provides methods to obtain components on the grid using the package gbasis
        """
        def __init__(self, mol, pbs_mol):

            self.mol   = mol
            self.basis = from_pyscf(mol)
            self.pbs   = from_pyscf(pbs_mol) if pbs_mol is not None else None
            self.atomic_charges = self.mol.atom_charges()
            self.atomic_coords  = self.mol.atom_coords()

            # Perform quick LDA calculation. Generates grid. 
            mf = dft.UKS(self.mol)
            mf.xc = 'svwn'
            mf.kernel()
            self.spherical_points = mf.grids().coords
            self.w                = mf.grids().weights
            self.mf = mf

            # # Build spherical grid using \textit{grid}
            # rad          = GaussLaguerre(70)
            # becke = BeckeWeights(order=3)
            # grid = MolGrid.from_preset( self.atomic_charges,
            #                             self.atomic_coords, 
            #                             rad,
            #                             ['fine' for i in range(len(self.atomic_charges))],
            #                             becke )
            
            # self.spherical_points = grid.points
            # self.w                = grid.weights

            # Build rectangular grid
            self.rectangular_grid   = None

        def assert_grid(self, grid):
            if grid == 'spherical':
                points = self.spherical_points
            elif grid == 'rectangular':
                assert self.rectangular_grid is not None, "Rectangular Grid must be defined first"
                points = self.rectangular_grid
            else: 
                raise ValueError("Specify either spherical or rectangular grid")
        
            return points

        def generate_grid(self, x, y, z):
            """
            Genrates Mesh from 3 separate linear spaces and flatten,
            needed for cubic grid.
            Parameters
            ----------
            grid: tuple of three np.ndarray
                (x, y, z)
            Returns
            -------
            grid: np.ndarray
                shape (3, len(x)*len(y)*len(z)).
            """
            # x,y,z, = grid
            shape = (len(x), len(y), len(z))
            X,Y,Z = np.meshgrid(x, y, z, indexing='ij')
            X = X.reshape((X.shape[0] * X.shape[1] * X.shape[2], 1))
            Y = Y.reshape((Y.shape[0] * Y.shape[1] * Y.shape[2], 1))
            Z = Z.reshape((Z.shape[0] * Z.shape[1] * Z.shape[2], 1))
            grid = np.concatenate((X,Y,Z), axis=1).T

            return grid, shape

        def build_rectangular(self, npoints):
            """
            Builds rectangular grid containing molecule

            Parameters
            ----------
            npoints: tuple
                Number of points per dimension (n_x, n_y, n_z)
            overage: float
                Spacial extent of box

            """

            # xmin, xmax = np.min(self.atomic_coords[:,0])+3, np.max(self.atomic_coords[:,0])+3
            # ymin, ymax = np.min(self.atomic_coords[:,1])+3, np.max(self.atomic_coords[:,0])+3
            # zmin, zmax = np.min(self.atomic_coords[:,2])+3, np.max(self.atomic_coords[:,0])+3
            
            g1 = np.linspace(-10, 10, npoints[0])
            g2 = np.linspace(0, 0, npoints[1])
            g3 = np.linspace(0, 0, npoints[2])
            gx, gy, gz = np.meshgrid(g1, g2, g3)
            g3d = np.vstack( [gx.ravel(), gy.ravel(), gz.ravel()] ).T

            self.x                = g1
            self.y                = g2
            self.z                = g3
            self.rectangular_grid = g3d

        def density(self, Da, Db=None, grid='spherical'):
            """
            Computes density on grid. 

            Parameters
            ----------

            density: np.ndarray.
                Density in AO basis

            grid: str.
                Type of grid used. Default spherical 
                If 'rectangular' used self.rectangular_grid != None 

            Returns
            -------
            density_g: np.ndarray
                Density on the requested grid    
            """

            points = self.assert_grid(grid)

            density_a = evaluate_density(Da, self.basis, points)
            if Db is not None:
                density_b = evaluate_density(Db, self.basis, points)
                density_g = np.concatenate([density_a, density_b])
                return density_g
            else:
                return density_a

        def hartree(self, density, grid='spherical'):
            """
            Computes Hartree Potential on grid. 

            Parameters
            ----------

            density: np.ndarray.
                Density in AO basis

            grid: str.
                Type of grid used. Default spherical 
                If 'rectangular' used self.rectangular_grid != None 


            Returns
            -------

            hartree_potential: np.ndarray
                Hartree potential on the requested grid
            """        
            points = self.assert_grid(grid)

            hartree_potential = point_charge_integral(self.basis, 
                                                    points, 
                                                    -np.ones(points.shape[0]), 
                                                    transform=None, 
                                                    coord_type='spherical')

            hartree_potential *= density[:, :, None]
            hartree_potential = np.sum(hartree_potential, axis=(0, 1))

            return hartree_potential

        def external(self, grid='spherical'):
            """
            Computes External Potential on grid. 

            Parameters
            ----------
            grid: str
                Type of grid used. Default spherical 
                If 'rectangular' used self.rectangular_grid != None

            Returns
            -------
            external_potential: np.ndarray
                External potential on the given grid. 
            """        
            points = self.assert_grid(grid)       

            old_settings = np.seterr(divide="ignore")  # silence warning for dividing by zero
            external_potential = self.atomic_charges[None, :] \
            / (np.sum((points[:, :, None] - self.atomic_coords.T[None, :, :]) ** 2, axis=1) ** 0.5)
            np.seterr(**old_settings)

            if external_potential.ndim > 1:
                external_potential = np.sum(external_potential, axis=1)

            return -external_potential

        def to_grid(self, f_nm, grid='spherical'):
            """
            Expresses a matrix quantity on the grid

            Parameters
            ----------
            coeff: np.ndarray
                Vector/Matrix on ao basis. 
                Shape: {(num_ao_basis, ), (num_ao_basis, num_ao_basis)}
            grid: str
                Type of grid used. Default spherical 
                If 'rectangular' used self.rectangular_grid != None

            Returns
            -------
            f_g: np.ndarray
                Vector/Matrix expressed on the requested grid
            """
            
            points = self.assert_grid(grid)

            if self.pbs is None:
                basis = self.basis
            else:
                basis = self.pbs

            phis = evaluate_basis(basis, points)
            f_g = f_nm.dot(phis)
            if f_nm.ndim == 2:
                f_g *= phis

            return f_g

        def to_ao(self, f_g, grid='spherical'):
            """
            Expresses grid quantity on the AO basis

            Parameters
            ----------
            f_g: np.ndarray
                Function expressed in g points 
            grid: str
                The grid used: 'radial' or 'spherical'

            Returns
            -------
            f_nm: np.ndarray
                f_g in ao basis
            """

            points = self.assert_grid(grid)

            phis = evaluate_basis(self.basis, points)
            f_nm = contract( 'pb, p,p,pa->ab', phis.T, f_g, self.w, phis.T )
            f_nm = 0.5 * (f_nm + f_nm.T)

            return f_nm
        
        def orbitals(self, C, grid='spherical'):
            """
            Obtains orbitals on grid

            Parameters
            ----------
            C: np.ndarray
                Molecular Orbitals on Atomic Orbital basis set. 
            grid: str
                The grid used: 'radial' or 'spherical'

            Returns
            -------
            mat_g: np.ndarray
                Orbitals in g points in space
            """

            points = self.assert_grid(grid)

            phis = evaluate_basis(self.basis, points)
            mat_g = C.T.dot(phis)
            return mat_g

        def laplacian_density(self, density, grid='spherical'):
            """
            Calculates the laplacian of the density
            
            Parameters
            ----------
            density: np.ndarray
                Density in the ao basis 
            grid: str
                The grid used: 'radial' or 'spherical'

            Returns
            -------
            lap_density: np.ndarray
                Laplacian of density given on g points in space
            """

            points = self.assert_grid(grid)

            lap_density = evaluate_density_laplacian( density, self.basis, points )
            return lap_density
    
        def gradient_density(self, density, grid='spherical'):        
            """
            Evaluatges gradient of density on requested grid

            Parameters
            ----------
            density: np.ndarray
                Density in the ao basis 
            grid: str
                The grid used: 'radial' or 'spherical'

            Returns
            -------
            grad_density: np.ndarray
                Gradient of density given on g points in space
            """

            points = self.assert_grid(grid)

            grad_density = evaluate_density_gradient(density, self.basis, points)
            return grad_density
            
        def ao_deriv(self, derivs=[0,0,0], transform=None, grid='spherical'):
            """ 
            Calculates AO on the grid (and its derivatives). 
            If Transformation is given, e.g. ao2mo, MO will be given

            Parameters
            ----------
            derivs: List of 3 integers
                Array that corresponds to the derivative of each spatial coordinate (x,y,z)
                0 -> no derivative
                1 -> first derivative ...
                [1,0,3] -> first derivative on x. 
                           no derivative on y.
                           third derivative on z

            transform: np.ndarray
                Matrix to transform within basis. E.g. ao2mo -> MO will be given.
            grid: str
                The grid used: 'radial' or 'spherical'

            Returns
            -------
            orbs_deriv: np.ndarray 
                Array of atomic orbitals and/or their derivatives. 
            """

            points = self.assert_grid(grid)

            orbs_deriv = evaluate_deriv_basis( self.basis, points, np.array(derivs), 
                                            transform=transform )

            return orbs_deriv

        # Specialized for methods. 
        # def posdef_kinetic_energy_density(self, density, grid='spherical'):
        #     """
        #     Please look at OuCarter or mRKS method
        #     Evaluates the positive-definite kinetic energy density on grid
        #     t = 1/2 \nabla \cdot \nabla \gamma(r,r') 
        #     """

        #     points = self.assert_grid(grid)
        #     t = evaluate_posdef_kinetic_energy_density(density, self.basis, points)

        #     return t

        # def kinetic_energy_density(self, density, alpha=-1/4, grid='spherical'):
        #     """
        #     Please look at OuCarter or mRKS method
        #     Evaluates the general form of the kinetic energy density
        #     t = 1/2 \nabla \cdot \nabla \gamma(r,r') + alpha \nabla^2 n(r)
        #     """

        #     points = self.assert_grid(grid)
        #     t = evaluate_general_kinetic_energy_density(density, self.basis, points, alpha=alpha)

        #     return t

        # def kinetic_energy_density_pauli(self, C, grid='spherical', method='grid'):
        #     """
        #     Please look at OuCarter or mRKS method
        #     Obtains kinetic energy density in terms of the Pauli kinetic energy density
            
        #     Parameters
        #     ----------
        #     C: np.ndarray
        #         Occupied Molecular Orbitals
        #     """

        #     points = self.assert_grid(grid)

        #     density = C @ C.T
        #     density_g = self.density(density, grid=grid)

        #     basis_dx = self.ao_deriv(derivs=[1,0,0], transform=None, grid=grid)
        #     basis_dy = self.ao_deriv(derivs=[0,1,0], transform=None, grid=grid)
        #     basis_dz = self.ao_deriv(derivs=[0,0,1], transform=None, grid=grid)

        #     if method == 'grid':
        #         orbs = self.orbitals(C, grid=grid)
        #         d_orbs = ((basis_dx + basis_dy + basis_dz).T @ C).T
        #         tau_p = np.zeros_like( density_g )
        #         for i in range(C.shape[1]):
        #             for j in range(C.shape[1]):
        #                 if i == j:
        #                     pass
        #                 else:
        #                     tau_p += np.abs( orbs[i,:] * (d_orbs[j,:]) - orbs[j,:] * (d_orbs[i,:]) )**2

        #     elif method == 'basis':
        #         basis = self.ao_deriv(grid=grid)
        #         dx = contract('pm,mi,nj,pn->ijp', basis.T, C, C, basis_dx.T)
        #         dy = contract('pm,mi,nj,pn->ijp', basis.T, C, C, basis_dy.T)
        #         dz = contract('pm,mi,nj,pn->ijp', basis.T, C, C, basis_dz.T)
            
        #         dx = (dx - np.transpose(dx, (1, 0, 2))) ** 2
        #         dy = (dy - np.transpose(dy, (1, 0, 2))) ** 2
        #         dz = (dz - np.transpose(dz, (1, 0, 2))) ** 2

        #         occ = np.ones(C.shape[1])
        #         occ_matrix = np.expand_dims(occ, axis=0) @ np.expand_dims(occ, axis=1)

        #         tau_p = np.sum((dx + dy + dz).T * occ_matrix, axis=(1,2)) 


        #     tau_p /= (2*density_g)

        #     return tau_p

        # def avg_local_orb_energy(self, density, orbitals, eigvals, grid='spherical'):
        #     """
        #     Please look at OuCarter or mRKS method
        #     Generates average local orbital energy. Described by Staroverov
        #     J. Chem. Phys. 146, 084103. [Equations 4 and/or 6]
        #     $$
        #     e_tilde = 1/n(r) * [ \sum_i \varepsilon_i * | \phi_i(r) |^2 ]
        #     $$
        #     """

        #     points = self.assert_grid(grid)

        #     phis      = evaluate_basis(self.basis, points)
        #     density_g = self.density(density=density, grid=grid)
        #     e_tilde   = contract('xp, xo, xo, x, xp-> p', phis, 
        #                                                 orbitals, orbitals, eigvals, 
        #                                                 phis) / density_g

        #     return e_tilde

        # def external_tilde(self, grid='spherical', method='grid'):
        #     """
        #     Please look at OuCarter or mRKS method
        #     Generates effective external potential from LDA exchange. Described by Ou + Carter. 
        #     J. Chem. Theory Comput. 2018, 14, 11, 5680–5689

        #     $$
        #     v^{~}{ext}(r) = \epsilon^{-LDA}(r) - \frac{\tau^{LDA}{L}}{n^{LDA}(r)}
        #     - v_{H}^{LDA}(r) - v_{xc}^{LDA}(r)
        #     $$
        #     (22) in [1].
        #     """

        #     points = self.assert_grid(grid)

        #     # LDA results
        #     Da0, Db0 = self.mf.make_rdm1()
        #     Ca0, Cb0 = self.mf.mo_coeff
        #     ea0, eb0 = self.mf.mo_energy

        #     da0_g = self.density(Da0, grid)
        #     db0_g = self.density(Db0, grid)

        #     # LDA exchange
        #     cx = -(3/np.pi)**(1/3)
        #     vxca = cx * da0_g ** (1/3)
        #     vxcb = cx * db0_g ** (1/3)

        #     # External tilde
        #     e_tilde = self.avg_local_orb_energy(Da0, Ca0, ea0, grid=grid)
        #     lap     = self.laplacian_density(Da0, grid=grid)
        #     grad    = self.gradient_density(Da0, grid=grid)
        #     grad    = grad[:,0] + grad[:,1] + grad[:,2]
        #     hartree = self.hartree(Da0, grid=grid)
            
        #     tau_l  = self.kinetic_energy_density_pauli(Ca0, grid=grid, method=method)
        #     tau_l += - 0.25 * lap + np.abs( grad )**2 / (8*da0_g) 
        
        #     external_tilde = e_tilde - tau_l/da0_g - hartree - vxca

        #     return external_tilde




"""
pdeco.py

Functions associated with PDE-Constrained Optimization.
"""

import numpy as np
from opt_einsum import contract
from scipy.optimize import minimize

class PDECO():
    """
    Performs Optimization as in: 10.1063/1.1535422 - Qin Wu + Weitao Yang

    Attributes:
    -----------
    lambda_rgl: {None, float}. If float, lambda-regularization is added with lambda=lambda_rgl.
    """

    regul_norm = None  # Regularization norm: ||v||^2
    lambda_reg = None  # Regularization constant

    def pdeco(self, opt_max_iter, reg=None, gtol=1e-3,
              opt_method='L-BFGS-B', opt=None):
        """
        Calls scipy minimizer to minimize lagrangian. 
        """
        self.lambda_reg = reg

        self.lambda_reg = reg
        if opt is None:
            opt = {"disp": False}
        opt['maxiter'] = opt_max_iter
        opt['gtol'] = gtol

        # Initialization for D and C
        self._diagonalize_with_potential_pbs(self.v_pbs)

        if self.S4 is None:
            self.S4 = self.eng.get_S4()

        if opt_method.lower() == 'bfgs' or opt_method.lower() == 'l-bfgs-b':
            opt_results = minimize( fun = self.lagrangian_pbeco,
                                    x0  = self.v_pbs, 
                                    jac = self.gradient_pbeco,
                                    method  = opt_method,
                                    options = opt
                                    )
        else:
            raise ValueError(F'{opt_method} is not supported. Only BFGS '
                             F'and L-BFGS is supported.')

        if opt_results.success == False:
            self.v_pbs = opt_results.x
            self.opt_info = opt_results
            raise ValueError("Optimization was unsucessful (|grad|=%.2e) within %i iterations, "
                             "try a different initial guess. %s"% (np.linalg.norm(opt_results.jac), opt_results.nit, opt_results.message)
                             )
        else:
            print(f"Optimization Successful within {opt_results.nit} iterations! |grad|={np.linalg.norm(opt_results.jac):.2e}." )
            self.v_pbs = opt_results.x
            self.opt_info = opt_results

    def lagrangian_pbeco(self, v):
        """
        Lagrangian to be minimized wrt external potential
        Equation (5) of main reference
        """

        # If v is not updated, will not re-calculate.
        if not np.allclose(v, self.v_pbs, atol=1e-15):
            self._diagonalize_with_potential_pbs(v)

        # self._diagonalize_with_potential_pbs(v)

        if self.ref == 1:
            L = 4 * contract("ijkl,ij,kl", self.S4, self.Da - self.Dt[0], self.Da- self.Dt[0])
        else:
            L = contract("ijkl,ij,kl", self.S4, self.Da+self.Db-self.Dt[0]-self.Dt[1], self.Da+self.Db-self.Dt[0]-self.Dt[1])
        # Add lambda-regularization
        if self.lambda_reg is not None:
            T = self.T_pbs
            if self.ref == 1:
                norm = 2 * (v[:self.npbs] @ T @ v[:self.npbs])
            else:
                norm = (v[self.npbs:] @ T @ v[self.npbs:]) + (v[:self.npbs] @ T @ v[:self.npbs])

            L += norm * self.lambda_reg
            self.regul_norm = norm
        return L

    def gradient_pbeco(self, v):
        """
        Calculates gradient wrt target density
        Equation (11) of main reference
        """
        # If v is not updated, will not re-calculate.
        if not np.allclose(v, self.v_pbs, atol=1e-15):
            self._diagonalize_with_potential_pbs(v)

        if self.ref == 1:
            grad_temp = np.zeros((self.nbf, self.nbf))
            g = 8 * contract("ijkl,ij,km->lm", self.S4, 2 * (self.Dt[0] - self.Da), self.Coca)  # shape (ao, mo)
            u = 0.5 * contract("lm,lm->m", self.Coca, g)  # shape (mo, )
            g -= 2 * contract('m,ij,jm->im', u, self.S2, self.Coca) # shape (ao, mo)
            for idx in range(self.nalpha):
                LHS = self.Fock - self.S2 * self.eigvecs_a[idx]
                p_i = np.linalg.solve(LHS, g[:, idx])
                # Gram–Schmidt rotation
                p_i -= np.sum(p_i * np.dot(self.S2, self.Coca[:,idx])) * self.Coca[:,idx]
                assert np.allclose([np.sum(p_i * (self.S2 @ self.Coca[:,idx])), np.linalg.norm(np.dot(LHS,p_i)-g[:, idx]), np.sum(g[:, idx]*self.Coca[:,idx])], 0, atol=1e-4)
                grad_temp += p_i[:, np.newaxis] * self.Coca[:,idx]

            self.grad = contract("ij,ijk->k", grad_temp, self.S3)
        else:
            grad_temp_a = np.zeros((self.nbf, self.nbf))
            g_a = 4 * contract("ijkl,ij,km->lm", self.S4, (self.Dt[0] - self.Da) + (self.Dt[1] - self.Db), self.Coca)  # shape (ao, mo)
            u_a = 0.5 * contract("lm,lm->m", self.Coca, g_a)  # shape (mo, )
            g_a -= 2 * contract('m,ij,jm->im', u_a, self.S2, self.Coca) # shape (ao, mo)
            for idx in range(self.nalpha):
                LHS = self.Fock[0] - self.S2 * self.eigvecs_a[idx]
                p_i = np.linalg.solve(LHS, g_a[:, idx])
                # Gram–Schmidt rotation
                p_i -= np.sum(p_i * np.dot(self.S2, self.Coca[:,idx])) * self.Coca[:,idx]
                assert np.allclose([np.sum(p_i * (self.S2 @ self.Coca[:,idx])), np.linalg.norm(np.dot(LHS,p_i)-g_a[:, idx]), np.sum(g_a[:, idx]*self.Coca[:,idx])], 0, atol=1e-4)
                grad_temp_a += p_i[:, np.newaxis] * self.Coca[:,idx]

            grad_temp_b = np.zeros((self.nbf, self.nbf))
            g_b = 4 * contract("ijkl,ij,km->lm", self.S4, (self.Dt[0] - self.Da) + (self.Dt[1] - self.Db), self.Cocb)  # shape (ao, mo)
            u_b = 0.5 * contract("lm,lm->m", self.Cocb, g_b)  # shape (mo, )
            g_b -= 2 * contract('m,ij,jm->im', u_b, self.S2, self.Cocb) # shape (ao, mo)
            for idx in range(self.nbeta):
                LHS = self.Fock[1] - self.S2 * self.eigvecs_b[idx]
                p_i = np.linalg.solve(LHS, g_b[:, idx])
                # Gram–Schmidt rotation
                p_i -= np.sum(p_i * np.dot(self.S2, self.Cocb[:,idx])) * self.Cocb[:,idx]
                assert np.allclose([np.sum(p_i * (self.S2 @ self.Cocb[:,idx])), np.linalg.norm(np.dot(LHS,p_i)-g_b[:, idx]), np.sum(g_b[:, idx]*self.Cocb[:,idx])], 0, atol=1e-4)
                grad_temp_b += p_i[:, np.newaxis] * self.Cocb[:,idx]

            self.grad = np.concatenate((contract("ij,ijk->k", grad_temp_a, self.S3), contract("ij,ijk->k", grad_temp_b, self.S3)))

        if self.lambda_reg is not None:
            T = self.T_pbs
            if self.ref == 1:
                rgl_vector = 4 * self.lambda_reg*np.dot(T, v[:self.npbs])
                self.grad += rgl_vector
            else:
                self.grad[:self.npbs] += 2 * self.lambda_reg*np.dot(T, v[:self.npbs])
                self.grad[self.npbs:] += 2 * self.lambda_reg*np.dot(T, v[self.npbs:])
        return self.grad

    def find_regularization_constant_pdeco(self, opt_max_iter, opt_method="L-BFGS-B", gtol=1e-3,
                                     opt=None, lambda_list=None):
        """
        Finding regularization constant lambda.

        Note: it is recommend to set a specific convergence criteria by opt or tol,
                in order to control the same convergence
                for different lambda value.

        After the calculation is done, one can plot the returns to select a good lambda.

        Parameters:
        -----------
        guide_potential_components: a list of string
            the components for guide potential v_pbs.
            see Inverter.generate_components() for details.

        opt_method: string default: "trust-krylov"
            opt_methods available in scipy.optimize.minimize

        tol: float
            Tolerance for termination. See scipy.optimize.minimize for details.

        opt: dictionary, optional
            if given:
                scipy.optimize.minimize(method=opt_method, options=opt)

        lambda_list: np.ndarray, optional
            A array of lambda to search; otherwise, it will be 10 ** np.linspace(-1, -7, 7).

        Returns:
        --------
        lambda_list: np.ndarray
            A array of lambda searched.

        P_list: np.ndarray
            The value defined by [Bulat, Heaton-Burgess, Cohen, Yang 2007] eqn (21).
            Corresponding to lambda in lambda_list.

        error_list: np.ndarray
            The Ts value for each lambda.


        """

        error_list = []
        v_norm_list = []


        if lambda_list is None:
            lambda_list = 10 ** np.linspace(-3, -9, 7)

        if opt is None:
            opt = {"disp"    : False}
        opt['maxiter'] = opt_max_iter
        opt['gtol'] = gtol

        self.lambda_reg = None
        # Initial calculation with no regularization
        # Initialization for D and C
        self._diagonalize_with_potential_pbs(self.v_pbs)
        if opt_method.lower() == 'bfgs' or opt_method.lower() == 'l-bfgs-b':
            initial_result = minimize(fun=self.lagrangian_pbeco,
                                   x0=self.v_pbs,
                                   jac=self.gradient_pbeco,
                                   method=opt_method,
                                   options=opt
                                   )
        else:
            raise ValueError(F'{opt_method} is not supported. Only BFGS '
                             F'and L-BFGS is supported.')

        if initial_result.success == False:
            raise ValueError("Optimization was unsucessful (|grad|=%.2e) within %i iterations, "
                             "try a different intitial guess"% (np.linalg.norm(initial_result.jac), initial_result.nit)
                             + initial_result.message)
        else:
            error0 = -initial_result.fun
            initial_v_pbs = initial_result.x  # This is used as the initial guess for with regularization calculation.

        for reg in lambda_list:
            self.lambda_reg = reg

            if opt_method.lower() == 'bfgs' or opt_method.lower() == 'l-bfgs-b':
                opt_results = minimize(fun=self.lagrangian_pbeco,
                                       x0=initial_v_pbs,
                                       jac=self.gradient_pbeco,
                                       method=opt_method,
                                       options=opt
                                       )

            else:
                raise ValueError(F'{opt_method} is not supported. Only BFGS '
                                 F'and L-BFGS is supported.')


            v_norm_list.append(self.regul_norm)
            error_list.append(opt_results.fun - self.lambda_reg * self.regul_norm)

        P_list = lambda_list * np.array(v_norm_list) / (np.array(error_list) - error0)

        return lambda_list, P_list, np.array(error_list)



"""
grider.py
"""

import numpy as np
from opt_einsum import contract
from pylibxc import LibXCFunctional as Functional

try:
    import psi4
    has_psi4 = True
except ImportError:
    has_psi4 = False

if has_psi4:
    class Psi4Grider():
        """
        Psi4 Grider
        Handles components on the grid using Psi4 grid blocks. 
        """
        def __init__(self, mol, basis, ref):
            
            self.mol = mol
            self.ref = ref
            self.basis = basis
            self.basis_str = basis.name()
            self.nbf   = self.basis.nbf()

            wfn_base = psi4.core.Wavefunction.build(self.mol, self.basis_str)
            self.wfn = psi4.proc.scf_wavefunction_factory('svwn', wfn_base, "UKS")
            self.wfn.initialize()

            self.nalpha = self.wfn.nalpha()
            self.nbeta  = self.wfn.nbeta()

            # Clean Vpot
            restricted = True if ref == 1 else False
            reference  = "RV" if ref == 1 else "UV"
            functional = psi4.driver.dft.build_superfunctional("SVWN", restricted=restricted)[0]
            self.Vpot = psi4.core.VBase.build(self.basis, functional, reference)
            self.Vpot.initialize()

            # self.spherical_points = np.array(self.Vpot.get_np_xyzw())[:3,:].T
            # self.w = np.array(self.Vpot.get_np_xyzw())[3,:].T

            self.npoints = self.Vpot.grid().npoints()   

        def grid_to_blocks(self, grid, basis=None):
            """
            Generate list of blocks to allocate given grid

            Parameters
            ----------
            grid: np.ndarray
                Grid to be distributed into blocks
                Size: (3, npoints) for homogeneous grid
                    (4, npoints) for inhomogenous grid to account for weights
            basis: psi4.core.BasisSet; optional
                The basis set. If not given, it will use target wfn.basisset().

            Returns
            -------
            blocks: list    
                List with psi4.core.BlockOPoints
            npoints: int
                Total number of points (for one dimension)
            points: psi4.core.{RKS, UKS}
                Points function to set matrices.
            """
            assert (grid.shape[0] == 3) or (grid.shape[0] == 4), """Grid does not have the correct dimensions. \n
                                                                Array must be of size (3, npoints) or (4, npoints)"""
            if_w = grid.shape[0] == 4

            if basis is None:
                basis = self.basis

            epsilon    = psi4.core.get_global_option("CUBIC_BASIS_TOLERANCE")
            extens     = psi4.core.BasisExtents(basis, epsilon)
            max_points = psi4.core.get_global_option("DFT_BLOCK_MAX_POINTS")        
            npoints    = grid.shape[1]
            nblocks = int(np.floor(npoints/max_points))
            blocks = []

            max_functions = 0
            #Run through full blocks
            idx = 0
            for nb in range(nblocks):
                x = psi4.core.Vector.from_array(grid[0][idx : idx + max_points])
                y = psi4.core.Vector.from_array(grid[1][idx : idx + max_points])
                z = psi4.core.Vector.from_array(grid[2][idx : idx + max_points])
                if if_w:
                    w = psi4.core.Vector.from_array(grid[3][idx : idx + max_points])
                else:
                    w = psi4.core.Vector.from_array(np.zeros(max_points))  # When w is not necessary and not given

                blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))
                idx += max_points
                max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
                                            else len(blocks[-1].functions_local_to_global())

            #Run through remaining points
            if idx < npoints:
                x = psi4.core.Vector.from_array(grid[0][idx:])
                y = psi4.core.Vector.from_array(grid[1][idx:])
                z = psi4.core.Vector.from_array(grid[2][idx:])
                if if_w:
                    w = psi4.core.Vector.from_array(grid[3][idx:])
                else:
                    w = psi4.core.Vector.from_array(np.zeros_like(grid[2][idx:]))  # When w is not necessary and not given
                blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))

                max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
                                            else len(blocks[-1].functions_local_to_global())

            zero_matrix = psi4.core.Matrix(basis.nbf(), basis.nbf())
            if self.ref == 1:
                point_func = psi4.core.RKSFunctions(basis, max_points, max_functions)
                point_func.set_pointers(zero_matrix)
            else:
                point_func = psi4.core.UKSFunctions(basis, max_points, max_functions)
                point_func.set_pointers(zero_matrix, zero_matrix)

            return blocks, npoints, point_func

        def generate_grid(self, x, y, z):
            """
            Genrates Mesh from 3 separate linear spaces and flatten,
            needed for cubic grid.
            Parameters
            ----------
            grid: tuple of three np.ndarray
                (x, y, z)
            Returns
            -------
            grid: np.ndarray
                shape (3, len(x)*len(y)*len(z)).
            """
            # x,y,z, = grid
            shape = (len(x), len(y), len(z))
            X,Y,Z = np.meshgrid(x, y, z, indexing='ij')
            X = X.reshape((X.shape[0] * X.shape[1] * X.shape[2], 1))
            Y = Y.reshape((Y.shape[0] * Y.shape[1] * Y.shape[2], 1))
            Z = Z.reshape((Z.shape[0] * Z.shape[1] * Z.shape[2], 1))
            grid = np.concatenate((X,Y,Z), axis=1).T

            return grid, shape

        def dft_grid_to_fock(self, value, Vpot=None):
                """For value on DFT spherical grid, Fock matrix is returned.
                VFock_ij = \int dx \phi_i(x) \phi_j(x) value(x)
                
                Parameters:
                -----------
                value: np.ndarray of shape (npoint, ).
                Vpot:psi4.core.VBase
                    Vpotential object with info about grid.
                    Provides DFT spherical grid. Only comes to play if no grid is given.
                
                Returns:
                ---------
                VFock: np.ndarray of shape (nbasis, nbasis)
                """
                if Vpot is None:
                    Vpot = self.Vpot
                VFock = np.zeros((self.nbf, self.nbf))
                points_func = Vpot.properties()[0]

                i = 0
                # Loop over the blocks
                for b in range(Vpot.nblocks()):
                    # Obtain block information
                    block = Vpot.get_block(b)
                    points_func.compute_points(block)
                    npoints = block.npoints()
                    lpos = np.array(block.functions_local_to_global())
                    if len(lpos) == 0:
                        i += npoints
                        continue
                    # Obtain the grid weight
                    w = np.array(block.w())

                    # Compute phi!
                    phi = np.array(points_func.basis_values()["PHI"])[:npoints, :lpos.shape[0]]

                    Vtmp = np.einsum('pb,p,p,pa->ab', phi, value[i:i+npoints], w, phi, optimize=True)

                    # Add the temporary back to the larger array by indexing, ensure it is symmetric
                    VFock[(lpos[:, None], lpos)] += 0.5 * (Vtmp + Vtmp.T)

                    i += npoints
                assert i == value.shape[0], "Did not run through all the points. %i %i" %(i, value.shape[0])
                return VFock

        def ao(self, coeff, grid=None, basis=None):
            """
            Generates a quantity on the grid given its ao representation.
            *This is the most general function for basis to grid transformation.

            Parameters
            ----------
            coeff: np.ndarray
                Vector/Matrix of quantity on ao basis. Shape: {(num_ao_basis, ), (num_ao_basis, num_ao_basis)}
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed.
            basis: psi4.core.BasisSet, optional
                The basis set. If not given it will use target wfn.basisset().
            Vpot: psi4.core.VBase
                Vpotential object with info about grid. 
                Provides DFT spherical grid. Only comes to play if no grid is given. 

            Returns
            -------
            coeff_r: np.ndarray Shape: (npoints, )
                Quantity expressed by the coefficient on the given grid 


            """


            if grid is not None:
                if type(grid) is np.ndarray:
                    if grid.shape[0] != 3 and grid.shape[0] != 4:
                        raise ValueError("The shape of grid should be (3, npoints) "
                                        "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                    blocks, npoints, points_function = self.grid_to_blocks(grid, basis=basis)
                else:
                    blocks, npoints, points_function = grid
            elif grid is None:
                Vpot = self.Vpot
                nblocks = Vpot.nblocks()
                blocks = [Vpot.get_block(i) for i in range(nblocks)]
                npoints = Vpot.grid().npoints()
                points_function = Vpot.properties()[0]
            else:
                raise ValueError("A grid or a V_potential (DFT grid) must be given.")

            coeff_r = np.zeros((npoints))

            offset = 0
            for i_block in blocks:
                points_function.compute_points(i_block)
                b_points = i_block.npoints()
                offset += b_points
                lpos = np.array(i_block.functions_local_to_global())
                if len(lpos)==0:
                    continue
                phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]

                if coeff.ndim == 1:
                    l_mat = coeff[(lpos[:])]
                    coeff_r[offset - b_points : offset] = contract('pm,m->p', phi, l_mat)
                elif coeff.ndim == 2:
                    l_mat = coeff[(lpos[:, None], lpos)]
                    coeff_r[offset - b_points : offset] = contract('pm,mn,pn->p', phi, l_mat, phi)

            return coeff_r

        def density(self, Da, Db=None, grid=None, Vpot=None):
            """
            Generates Density given grid

            Parameters
            ----------
            Da, Db: np.ndarray
                Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed.
            vpot: psi4.core.VBase
                Vpotential object with info about grid.
                Provides DFT spherical grid. Only comes to play if no grid is given. 

            Returns
            -------
            density: np.ndarray Shape: (ref, npoints)
                Density on the given grid. 
            """
            if Vpot is None:
                Vpot = self.Vpot

            Da = psi4.core.Matrix.from_array(Da)
            if Db is not None:
                Db = psi4.core.Matrix.from_array(Db)

            if grid is not None:
                if type(grid) is np.ndarray:
                    if grid.shape[0] != 3 and grid.shape[0] != 4:
                        raise ValueError("The shape of grid should be (3, npoints) "
                                        "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                    blocks, npoints, points_function = self.grid_to_blocks(grid)
                else:
                    blocks, npoints, points_function = grid
            elif grid is None:
                nblocks = Vpot.nblocks()
                blocks = [Vpot.get_block(i) for i in range(nblocks)]
                npoints = Vpot.grid().npoints()
                points_function = Vpot.properties()[0]
            else:
                raise ValueError("A grid or a V_potential (DFT grid) must be given.")

            if Db is None:
                points_function.set_pointers(Da)
                rho_a = points_function.point_values()["RHO_A"]
                density   = np.zeros((npoints))
            else:
                points_function.set_pointers(Da, Db)
                rho_a = points_function.point_values()["RHO_A"]
                rho_b = points_function.point_values()["RHO_B"]
                density   = np.zeros((npoints, self.ref))

            offset = 0
            for i_block in blocks:
                points_function.compute_points(i_block)
                b_points = i_block.npoints()
                offset += b_points

                if self.ref == 1:
                    density[offset - b_points : offset] = rho_a.np[ :b_points]
                else:
                    density[offset - b_points : offset, 0] = rho_a.np[ :b_points]
                    density[offset - b_points : offset, 1] = rho_b.np[ :b_points]

            return density

        def esp(self, Da=None, Db=None, vpot=None, grid=None, compute_hartree=True, wfn=None):
            """
            Generates EXTERNAL/ESP/HARTREE and Fermi Amaldi Potential on given grid
            Parameters
            ----------
            Da,Db: np.ndarray, opt, shape (nbf, nbf)
                The electron density in the denominator of Hartee potential. If None, the original density matrix
                will be used.
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed.
            vpot: psi4.core.VBase
                Vpotential object with info about grid.
                Provides DFT spherical grid. Only comes to play if no grid is given. 
            
            Returns
            -------
            vext, hartree, esp: np.ndarray
                External, Hartree, and Full Electronstatic potential on the given grid
                Shape: (npoints, )
            """
            if wfn is None:
                wfn = self.wfn

            if Da is not None or Db is not None:
                Da_temp = np.copy(self.wfn.Da().np)
                Db_temp = np.copy(self.wfn.Db().np)
                if Da is not None:
                    wfn.Da().np[:] = Da
                if Db is not None:
                    wfn.Db().np[:] = Db

            nthreads = psi4.get_num_threads()
            psi4.set_num_threads(1)
            
            if grid is not None:
                if type(grid) is np.ndarray:
                    blocks, npoints, points_function = self.grid_to_blocks(grid)
                else:
                    blocks, npoints, points_function = grid
            elif vpot is not None:
                nblocks = vpot.nblocks()
                blocks = [vpot.get_block(i) for i in range(nblocks)]
                npoints = vpot.grid().npoints()

            #Initialize Arrays
            vext = np.zeros(npoints)
            esp  = np.zeros(npoints)

            #Get Atomic Information
            mol_dict = self.mol.to_schema(dtype='psi4')
            natoms = len(mol_dict["elem"])
            indx = [i for i in range(natoms) if self.mol.charge(i) != 0.0]
            natoms = len(indx)
            #Atomic numbers and Atomic positions
            zs = [mol_dict["elez"][i] for i in indx]
            rs = [self.mol.geometry().np[i] for i in indx]

            if compute_hartree:
                esp_wfn = psi4.core.ESPPropCalc(wfn)

            #Loop Through blocks
            offset = 0
            with np.errstate(divide='ignore'):
                for i_block in blocks:
                    b_points = i_block.npoints()
                    offset += b_points
                    x = i_block.x().np
                    y = i_block.y().np
                    z = i_block.z().np

                    #EXTERNAL
                    for atom in range(natoms):
                        r =  np.sqrt((x-rs[atom][0])**2 + (y-rs[atom][1])**2 + (z-rs[atom][2])**2)
                        vext_temp = - 1.0 * zs[atom] / r
                        vext_temp[np.isinf(vext_temp)] = 0.0
                        vext[offset - b_points : offset] += vext_temp

                    if compute_hartree:
                        #ESP
                        xyz = np.concatenate((x[:,None],y[:,None],z[:,None]), axis=1)
                        grid_block = psi4.core.Matrix.from_array(xyz)
                        esp[offset - b_points : offset] = esp_wfn.compute_esp_over_grid_in_memory(grid_block).np

            #Hartree
            if compute_hartree:
                hartree = - 1.0 * (vext + esp)

            if Da is not None:
                self.wfn.Da().np[:] = Da_temp
            if Db is not None:
                self.wfn.Db().np[:] = Db_temp
            psi4.set_num_threads(nthreads)

            if compute_hartree:
                return vext, hartree, esp
            else:
                return vext

        def orbitals(self, Ca=None, Cb=None, grid=None, Vpot=None):
            """
            Generates orbitals given grid
            Parameters
            ----------
            Ca, Cb: np.ndarray
                Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed
            Vpot: psi4.core.VBase
                Vpotential object with info about grid.
                Provides DFT spherical grid. Only comes to play if no grid is given. 
            Returns
            -------
            orbitals: np.ndarray
                Orbitals on the given grid of size . 
                Shape: (nbasis, npoints, ref)
            """
            if Vpot is None:
                Vpot = self.Vpot

            if Ca is None and Cb is None:
                Ca = psi4.core.Matrix.from_array(self.Ca)
                Cb = psi4.core.Matrix.from_array(self.Cb)
            else:
                Ca = psi4.core.Matrix.from_array(Ca)
                Cb = psi4.core.Matrix.from_array(Cb)

            if self.ref == 2 and Cb is None:
                raise ValueError("Db is required for an unrestricted system")

            if grid is not None:
                if type(grid) is np.ndarray:
                    if grid.shape[0] != 3 and grid.shape[0] != 4:
                        raise ValueError("The shape of grid should be (3, npoints) "
                                        "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                    blocks, npoints, points_function = self.grid_to_blocks(grid)
                else:
                    blocks, npoints, points_function = grid
            elif grid is None and Vpot is not None:
                nblocks = Vpot.nblocks()
                blocks = [Vpot.get_block(i) for i in range(nblocks)]
                npoints = Vpot.grid().npoints()
                points_function = Vpot.properties()[0]
            else:
                raise ValueError("A grid or a V_potential (DFT grid) must be given.")

            if self.ref == 1:
                orbitals_r = [np.zeros((npoints)) for i_orb in range(self.nbf)]
                points_function.set_pointers(Ca)
                Ca_np = Ca.np
            if self.ref == 2:
                orbitals_r = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]
                points_function.set_pointers(Ca, Cb)
                Ca_np = Ca.np
                Cb_np = Cb.np

            offset = 0
            for i_block in blocks:
                points_function.compute_points(i_block)
                b_points = i_block.npoints()
                offset += b_points
                lpos = np.array(i_block.functions_local_to_global())
                if len(lpos)==0:
                    continue
                phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]

                for i_orb in range(self.nbf):
                    Ca_local = Ca_np[lpos, i_orb]
                    if self.ref == 1:
                        orbitals_r[i_orb][offset - b_points : offset] = contract('m, pm -> p', Ca_local, phi)
                    else:
                        Cb_local = Cb_np[lpos, i_orb]
                        orbitals_r[i_orb][offset - b_points : offset,0] = contract('m, pm -> p', Ca_local, phi)
                        orbitals_r[i_orb][offset - b_points : offset,1] = contract('m, pm -> p', Cb_local, phi)
            return orbitals_r

        def lap_phi(self, Ca=None, Cb=None, grid=None, Vpot=None):
            """
            Generates laplacian of molecular orbitals
            Parameters
            ----------
        Ca, Cb: np.ndarray
                Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed.
            Vpot: psi4.core.VBase
                Vpotential object with info about grid.
                Provides DFT spherical grid. Only comes to play if no grid is given. 
            Returns
            -------
            lap_phi: List[np.ndarray]. Where array is of shape (npoints, ref)
                Laplacian of molecular orbitals on the grid
            """
            if Vpot is None:
                Vpot = self.Vpot

            if Ca is None and Cb is None:
                Ca = psi4.core.Matrix.from_array(self.Ca)
                Cb = psi4.core.Matrix.from_array(self.Cb)
            else:
                Ca = psi4.core.Matrix.from_array(Ca)
                Cb = psi4.core.Matrix.from_array(Cb)

            if self.ref == 2 and Cb is None:
                raise ValueError("Db is required for an unrestricted system")

            if grid is not None:
                if type(grid) is np.ndarray:
                    if grid.shape[0] != 3 and grid.shape[0] != 4:
                        raise ValueError("The shape of grid should be (3, npoints) "
                                        "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                    blocks, npoints, points_function = self.grid_to_blocks(grid)
                else:
                    blocks, npoints, points_function = grid
            elif grid is None and Vpot is not None:
                nblocks = Vpot.nblocks()
                blocks = [Vpot.get_block(i) for i in range(nblocks)]
                npoints = Vpot.grid().npoints()
                points_function = Vpot.properties()[0]
            else:
                raise ValueError("A grid or a V_potential (DFT grid) must be given.")

            points_function.set_ansatz(2)

            if self.ref == 1:
                points_function.set_pointers(Ca)
                lap_phi = [np.zeros((npoints)) for i_orb in range(self.nbf)]
            else:
                points_function.set_pointers(Ca, Cb)
                lap_phi = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]

            offset = 0
            for i_block in blocks:
                points_function.compute_points(i_block)
                b_points = i_block.npoints()
                offset += b_points
                lpos = np.array(i_block.functions_local_to_global())
                if len(lpos)==0:
                    continue
                
                #Obtain subset of phi_@@ matrices
                lx = np.array(points_function.basis_values()["PHI_XX"])[:b_points, :lpos.shape[0]]
                ly = np.array(points_function.basis_values()["PHI_YY"])[:b_points, :lpos.shape[0]]
                lz = np.array(points_function.basis_values()["PHI_ZZ"])[:b_points, :lpos.shape[0]]

                for i_orb in range(self.nbf):
                    Ca_local = Ca.np[lpos, i_orb][:,None]
                    
                    if self.ref ==1:
                        lap_phi[i_orb][offset - b_points : offset] += ((lx + ly + lz) @ Ca_local)[:,0]
                    else:
                        Cb_local = Cb.np[lpos, i_orb][:,None]
                        lap_phi[i_orb][offset - b_points : offset, 0] += ((lx + ly + lz) @ Ca_local)[:,0]
                        lap_phi[i_orb][offset - b_points : offset, 1] += ((lx + ly + lz) @ Cb_local)[:,0]

            return lap_phi

        def grad_phi(self, Ca=None, Cb=None, grid=None,Vpot=None):
            """
            Generates laplacian of molecular orbitals
            Parameters
            ----------
        Ca, Cb: np.ndarray
                Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed.
            Vpot: psi4.core.VBase
                Vpotential object with info about grid.
                Provides DFT spherical grid. Only comes to play if no grid is given. 
            Returns
            -------
            grad_phi: List[np.ndarray]. Where array is of shape (npoints, ref)
                Gradient of molecular orbitals on the grid
            """
            if Vpot is None:
                Vpot = self.Vpot

            if Ca is None and Cb is None:
                Ca = psi4.core.Matrix.from_array(self.Ca)
                Cb = psi4.core.Matrix.from_array(self.Cb)
            else:
                Ca = psi4.core.Matrix.from_array(Ca)
                Cb = psi4.core.Matrix.from_array(Cb)

            if self.ref == 2 and Cb is None:
                raise ValueError("Db is required for an unrestricted system")

            if grid is not None:
                if type(grid) is np.ndarray:
                    if grid.shape[0] != 3 and grid.shape[0] != 4:
                        raise ValueError("The shape of grid should be (3, npoints) "
                                        "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
                    blocks, npoints, points_function = self.grid_to_blocks(grid)
                else:
                    blocks, npoints, points_function = grid
            elif grid is None and Vpot is not None:
                nblocks = Vpot.nblocks()
                blocks = [Vpot.get_block(i) for i in range(nblocks)]
                npoints = Vpot.grid().npoints()
                points_function = Vpot.properties()[0]
            else:
                raise ValueError("A grid or a V_potential (DFT grid) must be given.")

            points_function.set_ansatz(2)

            if self.ref == 1:
                points_function.set_pointers(Ca)
                grad_phi = [np.zeros((npoints)) for i_orb in range(self.nbf)]
            else:
                points_function.set_pointers(Ca, Cb)
                grad_phi = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]

            offset = 0
            for i_block in blocks:
                points_function.compute_points(i_block)
                b_points = i_block.npoints()
                offset += b_points
                lpos = np.array(i_block.functions_local_to_global())
                if len(lpos)==0:
                    continue
                
                #Obtain subset of phi_@ matrix
                gx = np.array(points_function.basis_values()["PHI_X"])[:b_points, :lpos.shape[0]]
                gy = np.array(points_function.basis_values()["PHI_Y"])[:b_points, :lpos.shape[0]]
                gz = np.array(points_function.basis_values()["PHI_Z"])[:b_points, :lpos.shape[0]]

                for i_orb in range(self.nbf):
                    Ca_local = Ca.np[lpos, i_orb][:,None]
                    if self.ref == 1:
                        grad_phi[i_orb][offset - b_points : offset] += ((gx + gy + gz) @ Ca_local)[:,0]
                    if self.ref == 2:
                        Cb_local = Cb.np[lpos, i_orb][:,None]
                        grad_phi[i_orb][offset - b_points : offset, 0] += ((gx + gy + gz) @ Ca_local)[:,0]
                        grad_phi[i_orb][offset - b_points : offset, 1] += ((gx + gy + gz) @ Cb_local)[:,0]

            return grad_phi

        def vxc(self, func_id=1, grid=None, Da=None, Db=None, Vpot=None):
            """
            Generates Vxc given grid
            Parameters
            ----------
            Da, Db: np.ndarray
                Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
            func_id: int
                Functional ID associated with Density Functional Approximationl.
                Full list of functionals: https://www.tddft.org/programs/libxc/functionals/
            grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
                grid where density will be computed.
            Vpot: psi4.core.VBase
                Vpotential object with info about grid.
                Provides DFT spherical grid. Only comes to play if no grid is given. 
            Returns
            -------
            VXC: np.ndarray
                Exchange correlation potential on the given grid
                Shape: (npoints, )
            """

            local_functionals = [1,546,549,532,692,641,552,287,307,578,5,24,4,579,308,289,551,
                                22,23,14,11,574,573,554,5900,12,13,25,9,10,27,3,684,683,17,7,
                                28,29,30,31,8,317,2,6,536,537,538,318,577,259,547,548,20,599,43,
                                51,580,50,550
                                ]

            if func_id not in local_functionals:
                raise ValueError("Only local functionals are supported on the grid")

            if Vpot is None:
                Vpot = self.Vpot

            if grid is not None:
                if type(grid) is np.ndarray:
                    blocks, npoints, points_function = self.grid_to_blocks(grid)
                else:
                    blocks, npoints, points_function = grid
                if self.ref == 1:
                    density = self.density(Da=Da, grid=grid)
                else:
                    density = self.density(Da=Da, Db=Db, grid=grid)
            elif grid is None and Vpot is not None:
                nblocks = Vpot.nblocks()
                blocks = [Vpot.get_block(i) for i in range(nblocks)]
                npoints = Vpot.grid().npoints()
                if self.ref == 1:
                    density = self.density(Da=Da),
                else:
                    density = self.density(Da=Da, Db=Db)
            else:
                raise ValueError("A grid or a V_potential (DFT grid) must be given.")

            ingredients = {'rho' : density}
            if self.ref == 1:
                functional = Functional(func_id, 1)
            else:
                functional = Functional(func_id, 2) 
            xc_dictionary = functional.compute(ingredients)
            vxc = xc_dictionary['vrho']

            return np.squeeze(vxc)

        # --------------------------------------------------------------------------

        def _average_local_orbital_energy(self, D, C, eig, grid_info=None, Vpot=None):
            """
            (4)(6) in mRKS.
            """
            if Vpot is None:
                Vpot = self.Vpot

            if grid_info is None:
                e_bar = np.zeros(Vpot.grid().npoints())
                nblocks = Vpot.nblocks()

                points_func = Vpot.properties()[0]
                points_func.set_deriv(0)
                blocks = None
            else:
                blocks, npoints, points_func = grid_info
                e_bar = np.zeros(npoints)
                nblocks = len(blocks)

                points_func.set_deriv(0)

            # For unrestricted
            iw = 0
            for l_block in range(nblocks):
                # Obtain general grid information
                if blocks is None:
                    l_grid = Vpot.get_block(l_block)
                else:
                    l_grid = blocks[l_block]
                l_npoints = l_grid.npoints()

                points_func.compute_points(l_grid)
                l_lpos = np.array(l_grid.functions_local_to_global())
                if len(l_lpos) == 0:
                    iw += l_npoints
                    continue
                l_phi = np.array(points_func.basis_values()["PHI"])[:l_npoints, :l_lpos.shape[0]]
                lD = D[(l_lpos[:, None], l_lpos)]
                lC = C[l_lpos, :]
                rho = contract('pm,mn,pn->p', l_phi, lD, l_phi)
                e_bar[iw:iw + l_npoints] = contract("pm,mi,ni,i,pn->p", l_phi, lC, lC, eig, l_phi) / rho

                iw += l_npoints
            assert iw == e_bar.shape[0], "Somehow the whole space is not fully integrated."
            return e_bar

        def _get_l_kinetic_energy_density_directly(self, D, C, grid_info=None, Vpot=None):
            """
            Calculate $\frac{\tau_L^{KS}}{\rho^{KS}}-\frac{\tau_P^{KS}}{\rho^{KS}}$:
            laplace_rho_temp: $\frac{\nabla^2 \rho}{4}$;
            tauW_temp: $\frac{|\napla \rho|^2}{8|\rho|}$;
            tauLmP_rho: $\frac{|\napla \rho|^2}{8|\rho|^2} - \frac{\nabla^2 \rho}{4\rho}$.

            (i.e. the 2dn and 3rd term in eqn. (17) in [1] over $\rho$.):
            """

            if Vpot is None:
                Vpot = self.Vpot

            if grid_info is None:
                tauLmP_rho = np.zeros(Vpot.grid().npoints())
                nblocks = Vpot.nblocks()
                points_func = Vpot.properties()[0]
                blocks = None

            else:
                blocks, npoints, points_func = grid_info
                tauLmP_rho = np.zeros(npoints)
                nblocks = len(blocks)

            points_func.set_deriv(2)

            iw = 0
            for l_block in range(nblocks):
                # Obtain general grid information
                if blocks is None:
                    l_grid = Vpot.get_block(l_block)
                else:
                    l_grid = blocks[l_block]
                l_npoints = l_grid.npoints()

                points_func.compute_points(l_grid)
                l_lpos = np.array(l_grid.functions_local_to_global())
                if len(l_lpos) == 0:
                    iw += l_npoints
                    continue
                l_phi = np.array(points_func.basis_values()["PHI"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_x = np.array(points_func.basis_values()["PHI_X"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_y = np.array(points_func.basis_values()["PHI_Y"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_z = np.array(points_func.basis_values()["PHI_Z"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_xx = np.array(points_func.basis_values()["PHI_XX"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_yy = np.array(points_func.basis_values()["PHI_YY"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_zz = np.array(points_func.basis_values()["PHI_ZZ"])[:l_npoints, :l_lpos.shape[0]]

                lD = D[(l_lpos[:, None], l_lpos)]
                # lC = C[l_lpos, :]

                rho = contract('pm,mn,pn->p', l_phi, lD, l_phi)
                rho_inv = 1/rho

                # Calculate the second term
                laplace_rho_temp = contract('ab,pa,pb->p', lD, l_phi, l_phi_xx + l_phi_yy + l_phi_zz)
                # laplace_rho_temp += contract('pm, mn, pn->p', l_phi_x,lD, l_phi_x)
                # laplace_rho_temp += contract('pm, mn, pn->p', l_phi_y,lD, l_phi_y)
                # laplace_rho_temp += contract('pm, mn, pn->p', l_phi_z,lD, l_phi_z)
                laplace_rho_temp += np.sum((l_phi_x @ lD) * l_phi_x, axis=1)
                laplace_rho_temp += np.sum((l_phi_y @ lD) * l_phi_y, axis=1)
                laplace_rho_temp += np.sum((l_phi_z @ lD) * l_phi_z, axis=1)

                laplace_rho_temp *= 0.25 * 2

                # Calculate the third term
                tauW_temp = contract('pm, mn, pn->p', l_phi, lD, l_phi_x) ** 2
                tauW_temp += contract('pm, mn, pn->p', l_phi, lD, l_phi_y) ** 2
                tauW_temp += contract('pm, mn, pn->p', l_phi, lD, l_phi_z) ** 2
                tauW_temp *= rho_inv * 0.125 * 4

                tauLmP_rho[iw: iw + l_npoints] = (-laplace_rho_temp + tauW_temp) * rho_inv
                iw += l_npoints
            assert iw == tauLmP_rho.shape[0], "Somehow the whole space is not fully integrated."

            return tauLmP_rho

        def _pauli_kinetic_energy_density(self, D, C, occ=None, grid_info=None, Vpot=None):
            """
            (16)(18) in mRKS. But notice this does not return taup but taup/n
            :return:
            """
            if Vpot is None:
                Vpot = self.Vpot

            if occ is None:
                occ = np.ones(C.shape[1])

            if grid_info is None:
                taup_rho = np.zeros(Vpot.grid().npoints())
                nblocks = Vpot.nblocks()

                points_func = Vpot.properties()[0]
                points_func.set_deriv(1)
                blocks = None

            else:
                blocks, npoints, points_func = grid_info
                taup_rho = np.zeros(npoints)
                nblocks = len(blocks)
                points_func.set_deriv(1)

            iw = 0
            for l_block in range(nblocks):
                # Obtain general grid information
                if blocks is None:
                    l_grid = Vpot.get_block(l_block)
                else:
                    l_grid = blocks[l_block]
                l_npoints = l_grid.npoints()

                points_func.compute_points(l_grid)
                l_lpos = np.array(l_grid.functions_local_to_global())
                if len(l_lpos) == 0:
                    iw += l_npoints
                    continue
                l_phi = np.array(points_func.basis_values()["PHI"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_x = np.array(points_func.basis_values()["PHI_X"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_y = np.array(points_func.basis_values()["PHI_Y"])[:l_npoints, :l_lpos.shape[0]]
                l_phi_z = np.array(points_func.basis_values()["PHI_Z"])[:l_npoints, :l_lpos.shape[0]]

                lD = D[(l_lpos[:, None], l_lpos)]

                rho = contract('pm,mn,pn->p', l_phi, lD, l_phi)

                lC = C[l_lpos, :]
                # Matrix Methods
                part_x = contract('pm,mi,nj,pn->ijp', l_phi, lC, lC, l_phi_x)
                part_y = contract('pm,mi,nj,pn->ijp', l_phi, lC, lC, l_phi_y)
                part_z = contract('pm,mi,nj,pn->ijp', l_phi, lC, lC, l_phi_z)
                part1_x = (part_x - np.transpose(part_x, (1, 0, 2))) ** 2
                part1_y = (part_y - np.transpose(part_y, (1, 0, 2))) ** 2
                part1_z = (part_z - np.transpose(part_z, (1, 0, 2))) ** 2


                occ_matrix = np.expand_dims(occ, axis=1) @ np.expand_dims(occ, axis=0)

                taup = np.sum((part1_x + part1_y + part1_z).T * occ_matrix, axis=(1,2)) * 0.5

                taup_rho[iw:iw + l_npoints] = taup / rho ** 2 * 0.5

                iw += l_npoints
            assert iw == taup_rho.shape[0], "Somehow the whole space is not fully integrated."
            return taup_rho




"""
oucarter.py

Functions associated with Ou-Carter inversion
"""

import numpy as np
import psi4

class OC():
    """
    Ou-Carter density to potential inversion [1].
    [1] [J. Chem. Theory Comput. 2018, 14, 5680−5689]
    """

    def _get_optimized_external_potential(self, grid_info, average_alpha_beta=False):
        """
        $
        v^{~}{ext}(r) = \epsilon^{-LDA}(r)
        - \frac{\tau^{LDA}{L}}{n^{LDA}(r)}
        - v_{H}^{LDA}(r) - v_{xc}^{LDA}(r)
        $
        (22) in [1].
        """

        Nalpha = self.nalpha
        Nbeta = self.nbeta

        # SVWN calculation
        wfn_LDA = psi4.energy("SVWN/" + self.eng.basis_str, molecule=self.eng.mol, return_wfn=True)[1]
        Da_LDA = wfn_LDA.Da().np
        Db_LDA = wfn_LDA.Db().np
        Ca_LDA = wfn_LDA.Ca().np
        Cb_LDA = wfn_LDA.Cb().np
        epsilon_a_LDA = wfn_LDA.epsilon_a().np
        epsilon_b_LDA = wfn_LDA.epsilon_b().np
        LDA_Vpot = wfn_LDA.V_potential()
        self.LDA_Vpot = LDA_Vpot

        vxc_LDA_DFT  = self.eng.grid.vxc(func_id=1, Da=Da_LDA, Db=Db_LDA, Vpot=LDA_Vpot) 
        vxc_LDA_DFT += self.eng.grid.vxc(func_id=8, Da=Da_LDA, Db=Db_LDA, Vpot=LDA_Vpot)
        vxc_LDA      = self.eng.grid.vxc(func_id=1,Da=Da_LDA, Db=Db_LDA, grid=grid_info) 
        vxc_LDA     += self.eng.grid.vxc(func_id=8,Da=Da_LDA, Db=Db_LDA, grid=grid_info)
        if self.ref != 1:
            assert vxc_LDA.shape[-1] == 2
            vxc_LDA_beta = vxc_LDA[:,1]
            vxc_LDA = vxc_LDA[:, 0]
            vxc_LDA_DFT_beta = vxc_LDA_DFT[:, 1]
            vxc_LDA_DFT = vxc_LDA_DFT[:, 0]

        # _average_local_orbital_energy() taken from mrks.py.
        e_bar_DFT      = self.eng.grid._average_local_orbital_energy(Da_LDA, Ca_LDA[:,:Nalpha], epsilon_a_LDA[:Nalpha], Vpot=LDA_Vpot)
        tauLmP_rho_DFT = self.eng.grid._get_l_kinetic_energy_density_directly(Da_LDA, Ca_LDA[:,:Nalpha], Vpot=LDA_Vpot)
        tauP_rho_DFT   = self.eng.grid._pauli_kinetic_energy_density(Da_LDA, Ca_LDA[:,:Nalpha], Vpot=LDA_Vpot)

        tauLmP_rho     = self.eng.grid._get_l_kinetic_energy_density_directly(Da_LDA, Ca_LDA[:,:Nalpha], grid_info=grid_info)
        e_bar          = self.eng.grid._average_local_orbital_energy(Da_LDA, Ca_LDA[:, :Nalpha], epsilon_a_LDA[:Nalpha], grid_info=grid_info)
        tauP_rho       = self.eng.grid._pauli_kinetic_energy_density(Da_LDA, Ca_LDA[:,:Nalpha], grid_info=grid_info)
        # print("taulp", tauLmP_rho)
        # print("Ebar", e_bar)
        # print("taup", tauP_rho)

        tauL_rho_DFT = tauLmP_rho_DFT + tauP_rho_DFT
        tauL_rho = tauLmP_rho + tauP_rho

        vext_opt_no_H_DFT = e_bar_DFT - tauL_rho_DFT - vxc_LDA_DFT
        vext_opt_no_H = e_bar - tauL_rho - vxc_LDA

        J = self.compute_hartree(Ca_LDA[:,:Nalpha],  Cb_LDA[:,:Nbeta])
        vext_opt_no_H_DFT_Fock = self.eng.grid.dft_grid_to_fock(vext_opt_no_H_DFT, LDA_Vpot)
        vext_opt_DFT_Fock = vext_opt_no_H_DFT_Fock - J[0] - J[1]

        # # Does vext_opt need a shift?
        # Fock_LDA = self.T + vext_opt_DFT_Fock + J[0] + J[1] + self.dft_grid_to_fock(vxc_LDA_DFT, self.Vpot)
        # eigvecs_a = self.diagonalize(Fock_LDA, self.nalpha)[-1]
        # shift = eigvecs_a[Nalpha-1] - epsilon_a_LDA[Nalpha-1]
        # vext_opt_DFT_Fock -= shift * self.S2
        # print("LDA shift:", shift, eigvecs_a[Nalpha-1], epsilon_a_LDA[Nalpha-1])

        vH = self.eng.grid.esp(grid=grid_info, wfn=wfn_LDA)[1]
        vext_opt = vext_opt_no_H - vH
        # vext_opt -= shift

        if self.ref != 1:
            e_bar_DFT_beta = self.eng.grid._average_local_orbital_energy(Db_LDA, Cb_LDA[:,:Nbeta], epsilon_b_LDA[:Nbeta], Vpot=LDA_Vpot)
            e_bar_beta = self.eng.grid._average_local_orbital_energy(Db_LDA, Cb_LDA[:, :Nbeta], epsilon_b_LDA[:Nbeta], grid_info=grid_info)

            tauLmP_rho_DFT_beta = self.eng.grid._get_l_kinetic_energy_density_directly(Db_LDA, Cb_LDA[:,:Nbeta], Vpot=LDA_Vpot)
            tauLmP_rho_beta = self.eng.grid._get_l_kinetic_energy_density_directly(Db_LDA, Cb_LDA[:,:Nalpha], grid_info=grid_info)

            tauP_rho_DFT_beta = self.eng.grid._pauli_kinetic_energy_density(Db_LDA, Cb_LDA[:,:Nbeta], Vpot=LDA_Vpot)
            tauP_rho_beta = self.eng.grid._pauli_kinetic_energy_density(Db_LDA, Cb_LDA[:,:Nbeta], grid_info=grid_info)

            tauL_rho_DFT_beta = tauLmP_rho_DFT_beta + tauP_rho_DFT_beta
            tauL_rho_beta = tauLmP_rho_beta + tauP_rho_beta

            vext_opt_no_H_DFT_beta = e_bar_DFT_beta - tauL_rho_DFT_beta - vxc_LDA_DFT_beta
            vext_opt_no_H_beta = e_bar_beta - tauL_rho_beta - vxc_LDA_beta

            vext_opt_no_H_DFT_Fock_beta = self.eng.grid.dft_grid_to_fock(vext_opt_no_H_DFT_beta, LDA_Vpot)
            vext_opt_DFT_Fock_beta = vext_opt_no_H_DFT_Fock_beta - J[0] - J[1]

            vext_opt_beta = vext_opt_no_H_beta - vH

            # vext_opt_DFT_Fock = (vext_opt_DFT_Fock + vext_opt_DFT_Fock_beta) * 0.5
            # vext_opt = (vext_opt + vext_opt_beta) * 0.5

            return (vext_opt_DFT_Fock, vext_opt_DFT_Fock_beta), (vext_opt, vext_opt_beta)
        return vext_opt_DFT_Fock, vext_opt
    
    def oucarter(self, maxiter, vxc_grid, D_tol=1e-7,
             eig_tol=1e-4, frac_old=0.5, init="scan"):
        """
        (23) in [1].
        [1] [J. Chem. Theory Comput. 2018, 14, 5680−5689]

        parameters:
        ----------------------
            maxiter: int
                same as opt_max_iter
            vxc_grid: np.ndarray of shape (3, num_grid_points)
                The final result will be represented on this grid
                default: 1e-4
            D_tol: float, opt
                convergence criteria for density matrices.
                default: 1e-7
            eig_tol: float, opt
                convergence criteria for occupied eigenvalue spectrum.
                default: 1e-4
            frac_old: float, opt
                Linear mixing parameter for current vxc and old vxc.
                If 0, no old vxc is mixed in.
                Should be in [0,1)
                default: 0.5.
            init: string, opt
                Initial guess method.
                default: "SCAN"
                1) If None, input wfn info will be used as initial guess.
                2) If "continue" is given, then it will not initialize
                but use the densities and orbitals stored. Meaningly,
                one can run a quick WY calculation as the initial
                guess. This can also be used to user speficified
                initial guess by setting Da, Coca, eigvec_a.
                3) If it's not continue, it would be expecting a
                method name string that works for psi4. A separate psi4 calculation
                would be performed.

        """
        
        # if self.ref != 1:
        #     raise ValueError("Currently only supports Spin-Restricted "
        #                      "calculations since Spin-Unrestricted CI "
        #                      "is not supported by Psi4.")

        grid_info = self.eng.grid.grid_to_blocks(vxc_grid) 
        if self.ref == 1:
            grid_info[-1].set_pointers(self.eng.wfn.Da())
            vext_opt_Fock, vext_opt = self._get_optimized_external_potential(grid_info)
        else:
            grid_info[-1].set_pointers(self.eng.wfn.Da(), self.eng.wfn.Db())
            (vext_opt_Fock, vext_opt_Fock_beta), (vext_opt, vext_opt_beta) = self._get_optimized_external_potential(grid_info)

        # Print Hasta aca vamos bien

        assert self.LDA_Vpot is not None        
        vH0_Fock = self.va

        Nalpha = self.nalpha
        Nbeta = self.nbeta
        # Initialization.
        if init is None:
            self.Da        = np.copy(self.Dt[0])
            self.Db        = np.copy(self.Dt[1])
            self.Coca      = np.copy(self.ct[0])
            self.Cocb      = np.copy(self.ct[1])
            self.eigvecs_a = self.eng.wfn.epsilon_a().np[:Nalpha]
            self.eigvecs_b = self.eng.wfn.epsilon_b().np[:Nbeta]
        elif init.lower()=="continue":
            pass
        else:
            wfn_temp = psi4.energy(init+"/" + self.eng.grid.basis_str, 
                                   molecule=self.eng.grid.mol, 
                                   return_wfn=True)[1]
            self.Da = np.array(wfn_temp.Da())
            self.Db = np.array(wfn_temp.Db())
            self.Coca = np.array(wfn_temp.Ca())[:, :Nalpha]
            self.Cocb = np.array(wfn_temp.Cb())[:, :Nbeta]
            self.eigvecs_a = np.array(wfn_temp.epsilon_a())
            self.eigvecs_b = np.array(wfn_temp.epsilon_b())
            del wfn_temp

        # nerror = self.on_grid_density(Da=self.Dt[0]-self.Da, Db=self.Dt[1]-self.Da, Vpot=self.Vpot)
        # w = self.Vpot.get_np_xyzw()[-1]
        # nerror = np.sum(np.abs(nerror.T) * w)
        # print("nerror", nerror)

        vxc_old      = 0.0
        vxc_old_beta = 0.0
        Da_old       = 0.0
        eig_old      = 0.0
        tauLmP_rho = self.eng.grid._get_l_kinetic_energy_density_directly(self.Dt[0], self.ct[0][:, :Nalpha])
        if self.ref != 1:
            tauLmP_rho_beta = self.eng.grid._get_l_kinetic_energy_density_directly(self.Dt[1], self.ct[1][:, :Nbeta])

        for OC_step in range(1, maxiter+1):
            tauP_rho = self.eng.grid._pauli_kinetic_energy_density(self.Da, self.Coca)
            e_bar    = self.eng.grid._average_local_orbital_energy(self.Da, self.Coca, self.eigvecs_a[:Nalpha])
            shift    = self.eigvecs_a[Nalpha - 1] - self.eng.wfn.epsilon_a().np[Nalpha - 1]
            vxc_extH = e_bar - tauLmP_rho - tauP_rho - shift

            if self.ref != 1:
                tauP_rho_beta = self.eng.grid._pauli_kinetic_energy_density(self.Db, self.Cocb)
                e_bar_beta    = self.eng.grid._average_local_orbital_energy(self.Db, self.Cocb, self.eigvecs_b[:Nbeta])
                shift_beta    = self.eigvecs_b[Nbeta - 1] - self.eng.wfn.epsilon_b().np[Nbeta - 1]
                vxc_extH_beta = e_bar_beta - tauLmP_rho_beta - tauP_rho_beta - shift_beta

            Derror = np.linalg.norm(self.Da - Da_old) / self.nbf ** 2
            eerror = (np.linalg.norm(self.eigvecs_a[:Nalpha] - eig_old) / Nalpha)
            if (Derror < D_tol) and (eerror < eig_tol):
                print("KSDFT stops updating.")
                break

            # linear Mixture
            if OC_step != 1:
                vxc_extH = vxc_extH * (1 - frac_old) + vxc_old * frac_old
                if self.ref != 1:
                    vxc_extH_beta = vxc_extH_beta * (1 - frac_old) + vxc_old_beta * frac_old

            vxc_old = np.copy(vxc_extH)
            if self.ref != 1:
                vxc_old_beta = np.copy(vxc_extH_beta)

            # Save old data.
            Da_old = np.copy(self.Da)
            eig_old = np.copy(self.eigvecs_a[:Nalpha])

            Vxc_extH_Fock = self.eng.grid.dft_grid_to_fock(vxc_extH, self.LDA_Vpot)
            Vxc_Fock      = Vxc_extH_Fock - vext_opt_Fock - vH0_Fock

            # print("1", np.diag(Vxc_extH_Fock))
            # print("2", np.diag(vext_opt_Fock))
            # print("3", np.diag(vH0_Fock))
            # print("4", np.diag(Vxc_Fock))

            if self.ref != 1:
                Vxc_extH_Fock_beta = self.eng.grid.dft_grid_to_fock(vxc_extH_beta, self.LDA_Vpot)
                Vxc_Fock_beta      = Vxc_extH_Fock_beta - vext_opt_Fock_beta - vH0_Fock

            if self.ref == 1:
                self._diagonalize_with_potential_vFock(v=Vxc_Fock)
            else:
                self._diagonalize_with_potential_vFock(v=(Vxc_Fock, Vxc_Fock_beta))

            print(f"Iter: {OC_step}, Density Change: {Derror:2.2e}, Eigenvalue Change: {eerror:2.2e}.")
            # nerror = self.on_grid_density(Da=self.Dt[0] - self.Da, Db=self.Dt[1] - self.Da, Vpot=self.Vpot)
            # nerror = np.sum(np.abs(nerror.T) * w)
            # print("nerror", nerror)

        # Calculate vxc on grid
        vH0        = self.eng.grid.esp(grid=grid_info, wfn=self.eng.wfn)[1]
        tauLmP_rho = self.eng.grid._get_l_kinetic_energy_density_directly(self.Dt[0], self.ct[0][:, :Nalpha], grid_info=grid_info)
        tauP_rho   = self.eng.grid._pauli_kinetic_energy_density(self.Da, self.Coca, grid_info=grid_info)
        e_bar      = self.eng.grid._average_local_orbital_energy(self.Da, self.Coca, self.eigvecs_a[:Nalpha], grid_info=grid_info)
        shift      = self.eigvecs_a[Nalpha - 1] - self.eng.wfn.epsilon_a().np[Nalpha - 1]

        if self.ref != 1:
            tauLmP_rho_beta = self.eng.grid._get_l_kinetic_energy_density_directly(self.Dt[1], self.ct[1][:, :Nbeta],grid_info=grid_info)
            tauP_rho_beta   = self.eng.grid._pauli_kinetic_energy_density(self.Db, self.Cocb, grid_info=grid_info)
            e_bar_beta      = self.eng.grid._average_local_orbital_energy(self.Db, self.Cocb, self.eigvecs_b[:Nbeta], grid_info=grid_info)
            shift_beta      = self.eigvecs_b[Nbeta - 1] - self.eng.wfn.epsilon_b().np[Nbeta - 1]

        if self.ref == 1:
            self.grid_vxc = e_bar - tauLmP_rho - tauP_rho - vext_opt - vH0 - shift
            return self.grid_vxc, e_bar, tauLmP_rho, tauP_rho, vext_opt, vH0, shift
        else:
            self.grid_vxc = np.array((e_bar - tauLmP_rho - tauP_rho - vext_opt - vH0 - shift,
                                      e_bar_beta - tauLmP_rho_beta - tauP_rho_beta - vext_opt_beta - vH0 - shift_beta
                                      ))
            return self.grid_vxc, (e_bar, e_bar_beta), (tauLmP_rho, tauLmP_rho_beta), \
                   (tauP_rho,tauP_rho_beta), (vext_opt, vext_opt_beta), vH0, (shift, shift_beta)
# """
# oucarter.py

# Functions associated with Ou-Carter inversion
# """

# import numpy as np
# from opt_einsum import contract
# import psi4

# class OC():
#     """
#     Ou-Carter density to potential inversion [1].
#     [1] [J. Chem. Theory Comput. 2018, 14, 5680−5689]
#     """
    
#     def oucarter(self, maxiter, D_tol=1e-7,
#              eig_tol=1e-4, frac_old=0.5, init="scan"):
#         """
#         (23) in [1].
#         [1] [J. Chem. Theory Comput. 2018, 14, 5680−5689]

#         parameters:
#         ----------------------
#             maxiter: int
#                 same as opt_max_iter
#             vxc_grid: np.ndarray of shape (3, num_grid_points)
#                 The final result will be represented on this grid
#                 default: 1e-4
#             D_tol: float, opt
#                 convergence criteria for density matrices.
#                 default: 1e-7
#             eig_tol: float, opt
#                 convergence criteria for occupied eigenvalue spectrum.
#                 default: 1e-4
#             frac_old: float, opt
#                 Linear mixing parameter for current vxc and old vxc.
#                 If 0, no old vxc is mixed in.
#                 Should be in [0,1)
#                 default: 0.5.
#             init: string, opt
#                 Initial guess method.
#                 default: "SCAN"
#                 1) If None, input wfn info will be used as initial guess.
#                 2) If "continue" is given, then it will not initialize
#                 but use the densities and orbitals stored. Meaningly,
#                 one can run a quick WY calculation as the initial
#                 guess. This can also be used to user speficified
#                 initial guess by setting Da, Coca, eigvec_a.
#                 3) If it's not continue, it would be expecting a
#                 method name string that works for psi4. A separate psi4 calculation
#                 would be performed.

#         """
        
#         if self.eng_str == 'psi4':
#             # if vxc_grid is not None:
#             #     plotting_grid = self.eng.grid.grid_to_blocks(vxc_grid)

#             # Set pointers for target density
#             points_func = self.eng.grid.Vpot.properties()[0]
#             if self.ref == 1:
#                 da_psi4 = psi4.core.Matrix.from_array(self.Dt[0])
#                 points_func.set_pointers(da_psi4)
#             else:
#                 da_psi4 = psi4.core.Matrix.from_array(self.Dt[0])
#                 db_psi4 = psi4.core.Matrix.from_array(self.Dt[1])
#                 points_func.set_pointers(da_psi4, db_psi4)

        
#         nalpha     = self.nalpha
#         nbeta      = self.nbeta
#         vext_tilde = self.eng.grid.external_tilde()          #vext_tilde_nm = ?
#         vH0_Fock   = self.va

#         # Get initial guess from the target calculation or new one.
#         if init is None:
#             self.Da = np.copy(self.Dt[0])
#             self.Db = np.copy(self.Dt[1])

#             self.Ca = np.copy(self.ct[0])
#             self.Cb = np.copy(self.ct[1])

#             self.eigvecs_a = self.et[0]
#             self.eigvecs_b = self.et[1]

#         elif init.lower()=="continue":
#             pass

#         else:
#             # Use engine to make a standard calculation
#             D, C, e = self.eng.run_single_point(self.eng.mol, self.eng.basis_str, init)
#             if self.ref == 1:
#                 self.Da, self.Db = D/2, D/2
#                 self.Cocca, self.Coccb = C[:,:self.nalpha], C[:,:self.nbeta] 
#                 self.Ca, self.Cb = C, C
#                 self.eigvecs_a, self.eigvecs_b = e, e

#             else:
#                 self.Da, self.Db = D[0], D[1]
#                 self.Ca, self.Cb = C[0][:,:nalpha], C[1][:,:nbeta]
#                 self.Cocca, self.Coccb = C[0][:,:nalpha], C[1][:,:nbeta]
#                 self.eigvecs_a, self.eigvecs_b = e[0], e[1]

#         # Ou Carter SCF 
#         vxc_eff_old      = 0.0
#         vxc_eff_old_beta = 0.0
#         Da_old           = np.zeros_like(self.Da)
#         eig_old          = np.zeros_like(self.eigvecs_a)

#         # Target fixed components
#         da0_g      = self.eng.grid.density(density=self.Da)
#         gra_da0_g  = self.eng.grid.gradient_density(density=self.Da)
#         gra_da0_g  = gra_da0_g[:,0] + gra_da0_g[:,1] + gra_da0_g[:,2]
#         lap_da0_g  = self.eng.grid.laplacian_density(density=self.Da)   

#         kinetic_energy_method = 'basis'

#         # Eq. (26) 1st and 2nd, 5th and 6th. They remain fixed
#         vxc_eff0_g    = 0.25 * (lap_da0_g/da0_g) - 0.125 * (np.abs(gra_da0_g)**2/np.abs(da0_g)**2)
#         vext_tilde_g  = self.eng.grid.external_tilde(method=kinetic_energy_method)
#         vhartree_nm   = self.va

#         # Begin SCF iterations
#         for iteration in range(1, maxiter+1):
#             tau_p   = self.eng.grid.kinetic_energy_density_pauli(self.Cocca, method=kinetic_energy_method)
#             e_tilde = self.eng.grid.avg_local_orb_energy(self.Da, self.Cocca, self.eigvecs_a)
#             shift   = self.eigvecs_a[nalpha-1] - self.et[0][nalpha-1]

#             vxc_eff_g = vxc_eff0_g + e_tilde - tau_p / da0_g - shift

#             if self.ref == 2:
#                 print("Todavia no perra")

#         # Linear Mixture 


#         # Add exact Hartree + Vext_tilde
#             vxc_eff_nm = self.eng.grid.to_ao(  vxc_eff_g - vext_tilde_g  )
#             vxc_eff_nm -= vhartree_nm[0]
            
#             if self.ref == 2:
#                 print("Todavia no perra")

#             H = self.eng.get_T() + self.eng.get_V() + vxc_eff_nm
#             self.Ca, self.Cocca, self.Da, self.eigvecs_a = self.eng.diagonalize( H, nalpha )           

#         # Convergence (?) 
#             d_error = np.linalg.norm( self.Da - Da_old )
#             e_error = np.linalg.norm( self.eigvecs_a - eig_old )

#             print(f"Iter: {iteration}, Density Change: {d_error:2.2e}, Eigenvalue Change: {e_error:2.2e} Shift Value:{shift}")

#             if ( d_error < D_tol ) and (e_error < eig_tol):
#                 print("SCF convergence achieved")
#                 break 

#             Da_old  = self.Da.copy()
#             eig_old = self.eigvecs_a.copy()

#         # Vxc on rectangular grid
#         #  Targets. They remain unchanged
#         da0_g      = self.eng.grid.density(density=self.Dt[0], grid='rectangular')
#         gra_da0_g  = self.eng.grid.gradient_density(density=self.Dt[0], grid='rectangular')
#         gra_da0_g  = gra_da0_g[:,0] + gra_da0_g[:,1] + gra_da0_g[:,2]
#         lap_da0_g  = self.eng.grid.laplacian_density(density=self.Dt[0], grid='rectangular')
#         vext_tilde_g  = self.eng.grid.external_tilde(grid='rectangular', method=kinetic_energy_method)
#         vhartree_g   = self.eng.grid.hartree(density=self.Dt[0], grid='rectangular')

#         # SCF dependant quantities
#         tau_p      = self.eng.grid.kinetic_energy_density_pauli(self.Ca, grid='rectangular', method=kinetic_energy_method)
#         e_tilde    = self.eng.grid.avg_local_orb_energy(self.Da, self.Ca, self.eigvecs_a, grid='rectangular')
#         vxc_eff0_g    = 0.25 * (lap_da0_g/da0_g) - 0.125 * (np.abs(gra_da0_g)**2/da0_g**2)



#         self.vxc0_g     = vxc_eff0_g
#         self.vhartree_g = vhartree_g
#         self.vext_g     = vext_tilde_g
#         self.tau_p      = tau_p/da0_g
#         self.e_tilde    = e_tilde

#         self.vxc_g = vxc_eff0_g + e_tilde - tau_p/da0_g - vext_tilde_g - vhartree_g



    




#         #     # nerror = self.on_grid_density(Da=self.Dt[0] - self.Da, Db=self.Dt[1] - self.Da, Vpot=self.Vpot)
#         #     # nerror = np.sum(np.abs(nerror.T) * w)
#         #     # print("nerror", nerror)

#         # # Calculate vxc on grid

#         # vH0 = self.on_grid_esp(grid=grid_info)[1]
#         # tauLmP_rho = self._get_l_kinetic_energy_density_directly(self.Dt[0], self.ct[0][:, :Nalpha], grid_info=grid_info)
#         # tauP_rho = self._pauli_kinetic_energy_density(self.Da, self.Coca, grid_info=grid_info)
#         # shift = self.eigvecs_a[Nalpha - 1] - self.wfn.epsilon_a().np[Nalpha - 1]
#         # e_bar = self._average_local_orbital_energy(self.Da, self.Coca, self.eigvecs_a[:Nalpha], grid_info=grid_info)
#         # if self.ref != 1:
#         #     tauLmP_rho_beta = self._get_l_kinetic_energy_density_directly(self.Dt[1], self.ct[1][:, :Nbeta],
#         #                                                              grid_info=grid_info)
#         #     tauP_rho_beta = self._pauli_kinetic_energy_density(self.Db, self.Cocb, grid_info=grid_info)
#         #     shift_beta = self.eigvecs_b[Nbeta - 1] - self.wfn.epsilon_b().np[Nbeta - 1]
#         #     e_bar_beta = self._average_local_orbital_energy(self.Db, self.Cocb, self.eigvecs_b[:Nbeta], grid_info=grid_info)

#         # if self.ref == 1:
#         #     self.grid.vxc = e_bar - tauLmP_rho - tauP_rho - vext_opt - vH0 - shift
#         #     return self.grid.vxc, e_bar, tauLmP_rho, tauP_rho, vext_opt, vH0, shift
#         # else:
#         #     self.grid.vxc = np.array((e_bar - tauLmP_rho - tauP_rho - vext_opt - vH0 - shift,
#         #                               e_bar_beta - tauLmP_rho_beta - tauP_rho_beta - vext_opt_beta - vH0 - shift_beta
#         #                               ))
#         #     return self.grid.vxc, (e_bar, e_bar_beta), (tauLmP_rho, tauLmP_rho_beta), \
#         #            (tauP_rho,tauP_rho_beta), (vext_opt, vext_opt_beta), vH0, (shift, shift_beta)


import n2v 
from pyscf import gto, scf, lib, dft, ao2mo, df
import pytest
import numpy as np

@pytest.fixture
def ine():
    # Define Molecule
    Ne = gto.M(atom = """
                    Ne
                    """,
    basis = 'aug-cc-pvtz')

    # Perform Calculation
    mf = Ne.KS()
    mf.xc = 'scan'
    mf.kernel()

    # Extract data for n2v. 
    da, db = mf.make_rdm1()/2, mf.make_rdm1()/2
    ca, cb = mf.mo_coeff[:,:Ne.nelec[0]], mf.mo_coeff[:, :Ne.nelec[1]]
    ea, eb = mf.mo_energy, mf.mo_energy

    # Initialize inverter object. 
    ine = n2v.Inverter( engine='pyscf' )

    ine.set_system( Ne, 'aug-cc-pvtz' )
    ine.Dt = [da, db]
    ine.ct = [ca, cb]
    ine.et = [ea, eb]

    return ine

def test_zmp(ine):

    ine.invert(method='zmp', guide_components='fermi_amaldi', lambda_list=[750], opt_max_iter=100, opt_tol=1e-5)

    assert np.isclose(  ine.eigvecs_a[:5].all(),  
                        np.array( [-30.70166724,  -1.58399267,  -0.73369545,  -0.73369545, -0.73369545] ).all() ) 
    

def test_wuyang(ine):
    
    ine.invert("WuYang", guide_components="fermi_amaldi")

    assert np.isclose( ine.eigvecs_a[:5].all(),  
                        np.array( [-30.75604341,  -1.60572171,  -0.75614577,  -0.75614564, -0.75614546] ).all() )


def test_pedeco(ine):

    ine.invert("PDECO", opt_max_iter=200, guide_components="fermi_amaldi", gtol=1e-6)

    assert np.isclose(  ine.eigvecs_a[:5].all(),  
                        np.array( [-30.75843463,  -1.60691725,  -0.75708326,  -0.75708173, -0.75708055] ).all())

def test_oucarter(ine):

    x = np.linspace(-5, 10, 1501)
    y = [0]
    z = [0]
    grid, shape = ine.eng.grid.generate_grid(x, y, z)
    
    with pytest.raises(ValueError):
        ine.invert('OC', vxc_grid=grid)


def test_oucarter(ine):

    x = np.linspace(-5, 10, 1501)
    y = [0]
    z = [0]
    grid, shape = ine.eng.grid.generate_grid(x, y, z)
    
    with pytest.raises(ValueError):
        ine.invert('mRKS', vxc_grid=grid, opt_max_iter=30, frac_old=0.8, init='scan')